#include "stdafx.h"
#include "ArchiverDll.h"
#include "../Windows/Dll.h"
#include "../String/Encoding.h"
namespace Themis {

namespace {
    typedef void* (WINAPI* FnOpenArchive)(HWND, LPCSTR, DWORD);
    typedef int (WINAPI* FnFindFirst)(void*, LPCSTR, Detail::INDIVIDUALINFO*);
    typedef int (WINAPI* FnCloseArchive)(void*);
    typedef int (WINAPI* Fn)(HWND, LPCSTR, LPSTR , DWORD);
    typedef int (WINAPI* FnExtractMem)(HWND, LPCSTR, LPBYTE, DWORD, time_t*, LPWORD, LPDWORD);
    typedef BOOL (WINAPI* FnSetCursorMode)(BOOL);
    typedef BOOL (WINAPI* FnSetBackGroundMode)(BOOL);
    typedef BOOL (WINAPI* FnGetRunning)();
    typedef BOOL (WINAPI* FnConfigDialog)(HWND, LPSTR, int);
    typedef BOOL (WINAPI* FnSetOwner)(HWND, CArchiverDll::FnArchiverProc);
    typedef BOOL (WINAPI* FnCheckArchive)(LPCSTR, int);
    typedef BOOL (WINAPI* FnGetArcCompressedSizeEx)(void*, Detail::ULHA_INT64*);
    typedef int (WINAPI* FnGetMethod)(void*, LPSTR szBuffer, int nSize);
}

CArchiverDll::CArchiverDll(const string& strDllName, const string& strFuncPrefix) : m_strDllName(strDllName), m_strFuncPrefix(strFuncPrefix) {
    try {
        m_pDll.reset(new CDll(strDllName));
        mutex.reset(new boost::recursive_mutex);
        SetCursorMode(false);
        SetBackGroundMode(false);
    } catch (std::exception&) {
        m_pDll.reset();
        mutex.reset();
        Themis_DebugWarning(strDllName + L" ̓ǂݍ݂܂͓샂[h̐ݒɎs");
    }
}

CArchiverDll::~CArchiverDll() {
}

void CArchiverDll::SetCursorMode(bool b) {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnSetCursorMode fnSetCursorMode = static_cast<FnSetCursorMode>(m_pDll->GetProc(m_strFuncPrefix + L"SetCursorMode"));
            fnSetCursorMode(b ? TRUE : FALSE);
        } catch (EntryPointNotFoundException&) {
        }
    }
}

void CArchiverDll::SetBackGroundMode(bool b) {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnSetBackGroundMode fnSetBackGroundMode = static_cast<FnSetBackGroundMode>(m_pDll->GetProc(m_strFuncPrefix + L"SetBackGroundMode"));
            fnSetBackGroundMode(b ? TRUE : FALSE);
        } catch (EntryPointNotFoundException&) {
        }
    }
}

bool CArchiverDll::IsRunning() {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnGetRunning fnGetRunning = static_cast<FnGetRunning>(m_pDll->GetProc(m_strFuncPrefix + L"GetRunning"));
            return fnGetRunning() != FALSE;
        } catch (EntryPointNotFoundException&) {
        }
    }
    return false;
}

bool CArchiverDll::WaitForEndRunning(int nMaxSleep) {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnGetRunning fnGetRunning = static_cast<FnGetRunning>(m_pDll->GetProc(m_strFuncPrefix + L"GetRunning"));
            for (int i = 0; i < nMaxSleep; i++) {
                if (fnGetRunning() == FALSE) {
                    return true;
                } else {
                    ::Sleep(100);
                }
            }
        } catch (EntryPointNotFoundException&) {
        }
    }
    return false;
}

bool CArchiverDll::CheckArchive(const string& strFileName, int nMode) {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnCheckArchive fnCheckArchive = static_cast<FnCheckArchive>(m_pDll->GetProc(m_strFuncPrefix + L"CheckArchive"));
            if (fnCheckArchive(Encoding::GetDefault().GetBytes(strFileName).c_str(), nMode) != FALSE) {
                return true;
            }
        } catch (EntryPointNotFoundException&) {
        }
    }
    return false;
}

std::tr1::shared_ptr<CArchiverDll::CFindFile> CArchiverDll::FindFile(HWND hWnd, const string& strFileName, DWORD dwMode) {
    if (IsReady()) {
        try {
            boost::recursive_mutex::scoped_lock lock(*mutex);
            FnOpenArchive fnOpenArchive = static_cast<FnOpenArchive>(m_pDll->GetProc(m_strFuncPrefix + L"OpenArchive"));
            FnFindFirst fnFindFirst = static_cast<FnFindFirst>(m_pDll->GetProc(m_strFuncPrefix + L"FindFirst"));
            void* fnFindNext = m_pDll->GetProc(m_strFuncPrefix + L"FindNext");
            void* fnCloseArchive = m_pDll->GetProc(m_strFuncPrefix + L"CloseArchive");
            void* hArc = fnOpenArchive(hWnd, Encoding::GetDefault().GetBytes(strFileName).c_str(), dwMode);
            if (hArc != NULL) {
                Detail::INDIVIDUALINFO info;
                if (fnFindFirst(hArc, "*", &info) == 0) {
                    return std::tr1::shared_ptr<CArchiverDll::CFindFile>(
                        new CFindFile(*mutex, hArc, info, fnFindNext, fnCloseArchive));
                }
            }
        } catch (EntryPointNotFoundException&) {
        }
    }
    return std::tr1::shared_ptr<CFindFile>();
}

bool CArchiverDll::GetArcCompressedSizeEx(CFindFile& findFile, Detail::ULHA_INT64& size) {
    if (IsReady()) {
        try {
            FnGetArcCompressedSizeEx fn = static_cast<FnGetArcCompressedSizeEx>(m_pDll->GetProc(m_strFuncPrefix + L"GetArcCompressedSizeEx"));
            return fn(findFile.m_hArc, &size) != FALSE;
        } catch (EntryPointNotFoundException&) {
        }
    }
    return false;
}

/// k@̎擾
string CArchiverDll::GetMethod(CFindFile& findFile) {
    char sz[0x400] = {0};
    if (IsReady()) {
        try {
            FnGetMethod fn = static_cast<FnGetMethod>(m_pDll->GetProc(m_strFuncPrefix + L"GetMethod"));
            if (fn == NULL || fn(findFile.m_hArc, sz, sizeof(sz) - 1) != 0) {
                throw RuntimeException(L"k@̎擾Ɏs܂");
            }
        } catch (EntryPointNotFoundException&) {
        }
    }
    return Encoding::GetDefault().GetString(sz);
}

CArchiverDll::CFindFile::CFindFile(boost::recursive_mutex& e, void* hArc, Detail::INDIVIDUALINFO& info, void* fnNext, void* fnClose)
: m_lock(e), m_info(info), m_hArc(hArc), m_fnNext(fnNext), m_fnClose(fnClose) {
}

CArchiverDll::CFindFile::~CFindFile() {
    int nRet = static_cast<FnCloseArchive>(m_fnClose)(m_hArc);
    Themis_DebugAssert(nRet == 0);
}

void CArchiverDll::Command(HWND hWnd, const string& strCmd) {
    if (!IsReady()) {
        throw RuntimeException(m_strDllName + L" ǂݍݏo܂ł");
    }
    Themis_DebugLog(m_strDllName + L": " + strCmd);
    boost::recursive_mutex::scoped_lock lock(*mutex);
    Fn fn = static_cast<Fn>(m_pDll->GetProc(m_strFuncPrefix));
    char szBuf[0x400] = {0};
    int nRet = fn(hWnd, Encoding::GetDefault().GetBytes(strCmd).c_str(), szBuf, sizeof(szBuf) - 1);
    if (nRet != 0) {
        // sۂ
        if (szBuf[0] != L'\0') {
            throw RuntimeException(L"ɉ𓀎sF" + Encoding::GetDefault().GetString(szBuf));
        } else {
            throw RuntimeException(L"ɓWJɃG[܂");
        }
    }
}

bool CArchiverDll::ExtractMem(HWND hWnd, const string& strCmd, void* p, size_t n, time_t* pTime, WORD* pwAttr, DWORD* pdwWriteSize) {
    if (!IsReady()) {
        throw RuntimeException(m_strDllName + L" ǂݍݏo܂ł");
    }
    boost::recursive_mutex::scoped_lock lock(*mutex);
    FnExtractMem fnExtractMem = static_cast<FnExtractMem>(m_pDll->GetProc(m_strFuncPrefix + L"ExtractMem"));
    return fnExtractMem(hWnd, Encoding::GetDefault().GetBytes(strCmd).c_str(), LPBYTE(p), DWORD(n), pTime, pwAttr, pdwWriteSize) == 0;
}

void CArchiverDll::ConfigDialog(HWND hWnd, char* pszCmdBuf) {
    if (IsReady()) {
        boost::recursive_mutex::scoped_lock lock(*mutex);
        FnConfigDialog ConfigDialog = static_cast<FnConfigDialog>(m_pDll->GetProc(m_strFuncPrefix + L"ConfigDialog"));
        ConfigDialog(hWnd, pszCmdBuf, 1); // UNPACK_CONFIG_MODE
    }
}

void CArchiverDll::SetOwner(HWND hWnd, FnArchiverProc fn) {
    if (IsReady()) {
        boost::recursive_mutex::scoped_lock lock(*mutex);
        FnSetOwner SetOwner = static_cast<FnSetOwner>(m_pDll->GetProc(m_strFuncPrefix + L"SetOwnerWindowEx"));
        SetOwner(hWnd, fn);
    }
}

//////////////////////////////////////////////////////////////////////

CArchiverDll& CArchiverDll::GetUnZip() {
    static CArchiverDll s_oDll(L"UNZIP32.DLL", L"UnZip");
    return s_oDll;
}

CArchiverDll& CArchiverDll::GetUnLha() {
    static CArchiverDll s_oDll(L"UNLHA32.DLL", L"Unlha");
    return s_oDll;
}

CArchiverDll& CArchiverDll::GetUnRar() {
    static CArchiverDll s_oDll(L"unrar32.dll", L"Unrar");
    return s_oDll;
}

CArchiverDll& CArchiverDll::Get7Zip() {
    static CArchiverDll s_oDll(L"7-zip32.dll", L"SevenZip");
    return s_oDll;
}

CArchiverDll& CArchiverDll::GetUnGca() {
    static CArchiverDll s_oDll(L"UnGCA32.dll", L"UnGCA");
    return s_oDll;
}

}
