#pragma once
namespace Themis {

/// zip̈Ís
class CZipDecrypter {
    unsigned keys[3];
    static unsigned crc32tab[0x100];
    static bool crc32Initialized;
public:

    // pX[h̃`FbNDecrypter̐
    static auto_ptr<CZipDecrypter> Create(const string& password, const uint8_t* decryptedHeader, uint16_t cryptTest) {
        auto_ptr<CZipDecrypter> p(new CZipDecrypter(password));
        uint8_t hh[12];
        for (size_t i = 0; i < sizeof(hh); i++) {
            hh[i] = p->decode(decryptedHeader[i]);
        }
        uint16_t c = hh[12 - 2];
        uint16_t b = hh[12 - 1];
        if (uint16_t(c | b << 8) != cryptTest) {
            return auto_ptr<CZipDecrypter>(); // s
        }
        return p;
    }

    CZipDecrypter(const string& password) {
        // crcZope[ȕ
        if (!crc32Initialized) {
            const unsigned POLYNOMIAL = 0xedb88320;
            for (int i = 0; i < 0x100; i++) {
                unsigned c = i;
                for (int j = 0; j < 8; j++) {
                    c = (c & 1) != 0 ? (c >> 1) ^ POLYNOMIAL : c >> 1;
                }
                crc32tab[i] = c;
            }
            // [
            crc32Initialized = true;
        }
        // L[̍XV
        keys[0] = 0x12345678;
        keys[1] = 0x23456789;
        keys[2] = 0x34567890;
        for (int i = 0, n = password.GetLength(); i < n; i++) {
            update_keys((int)password[i]);
        }
    }

    inline uint8_t decode(uint8_t c) {
        c ^= decrypt_byte();
        update_keys(c);
        return c;
    }

private:
    inline uint8_t decrypt_byte() {
        unsigned t = (keys[2] & 0xffff) | 2;
        return uint8_t((t * (t ^ 1)) >> 8);
    }

    inline void update_keys(uint8_t c) {
        keys[0] = CRC32(keys[0], c);
        keys[1] += keys[0] & 0xff;
        keys[1] = keys[1] * 0x8088405 + 1;
        keys[2] = CRC32(keys[2], uint8_t(keys[1] >> 24));
    }

    inline unsigned CRC32(unsigned c, uint8_t b) const {
        return crc32tab[int(c ^ b) & 0xff] ^ (c >> 8);
    }
};

}
