#pragma once
#include <math.h>
namespace Themis {

/**
    𐔂IuWFNg̊NXB
*/
class ICounter : private boost::noncopyable {
public:
    virtual ~ICounter() {}
    virtual int Get() const = 0;
    virtual void Inc() = 0;
};

/**
    ŏlEőlEJE^B

    𕉒lɂInc()ɌĂ悤ɂB
    őlorŏl܂ōs炻ꂼŏlEől֖߂B
*/
class CIncrementalCounter : public ICounter {
public:
    CIncrementalCounter(int nMin, int nMax, int nStep=1)
        : m_nMin(nMin), m_nMax(nMax), m_nStep(nStep), m_n(0<nStep ? nMin : nMax) {}
    ~CIncrementalCounter() {}
    pair<int, int> GetRange() const { return make_pair(m_nMin, m_nMax); }
    int GetStep() const { return m_nStep; }
    int Get() const { return m_n; }
    void Set(int n) { m_n = n; }
    void Inc() {
        m_n += m_nStep;
        if (m_n < m_nMin || m_nMax < m_n) {         //nMin/nMax̒lĂ
            m_n = (0 < m_nStep) ? m_nMin : m_nMax;  //̕ɉnMin/nMax̒lB
        }
    }
private:
    int     m_nMin;
    int     m_nMax;
    int     m_nStep;
    int     m_n;
};

/**
    ŏlEől̊Ԃオ艺肷JE^B

    CIncrementalCounter̃Rsy
*/
class CRepeatCounter : public ICounter {
public:
    CRepeatCounter(int nMin, int nMax, int nStep=1)
        : m_nMin(nMin), m_nMax(nMax), m_nStep(nStep), m_n(0<nStep ? nMin : nMax) {}
    ~CRepeatCounter() {}
    pair<int, int> GetRange() const { return make_pair(m_nMin, m_nMax); }
    int GetStep() const { return m_nStep; }
    int Get() const { return m_n; }
    void Set(int n) { m_n = n; }
    void Inc() {
        m_n += m_nStep;
        if (m_n < m_nMin || m_nMax < m_n) {         //nMin/nMax̒lĂ
            m_n = (0 < m_nStep) ? m_nMax : m_nMin;  //̕ɉnMin/nMax̒lB
            m_nStep = - m_nStep;
        }
    }
private:
    int     m_nMin;
    int     m_nMax;
    int     m_nStep;
    int     m_n;
};

/**
    JnlIl܂ŁAȂ߂ÂĂJE^B

    XN[ƂɎgCB
*/
class CAccelCounter : public ICounter {
public:
    CAccelCounter(int nNow, int nEnd)
        : m_nNow(nNow), m_nEnd(nEnd), m_nSpeed(nNow<nEnd?1:-1) { m_nSpeedStep = m_nSpeed; }
    int GetEnd() const { return m_nEnd; }
    int Get() const { return m_nNow; }
    void Set(int n) { m_nNow = n; }
    void Inc() { m_nNow += m_nSpeed; m_nSpeed += m_nSpeedStep; }
private:
    int     m_nNow;
    int     m_nEnd;
    int     m_nSpeed;
    int     m_nSpeedStep;
};

/**
    JnlIl܂ŁAȂ߂ÂĂJE^B

    XN[ƂɎgCB
*/
class CDecelCounter : public ICounter {
public:
    CDecelCounter(int nNow, int nEnd)
        : m_nNow(nNow), m_nEnd(nEnd) {}
    int GetEnd() const { return m_nEnd; }
    int Get() const { return m_nNow; }
    void Set(int n) { m_nNow = n; }
    void Inc() { m_nNow += (m_nEnd - m_nNow + (m_nNow<m_nEnd?+1:-1)) / 2; }
private:
    int     m_nNow;
    int     m_nEnd;
};

/**
    x[Xl}U܂ŏオ艺肷JE^B

    cycléAx[Xlx[Xl{UAx[XlA
    x[Xl|UƂȂāAOxx[Xlɖ߂܂łInc()̉
*/
class CWaveCounter : public ICounter {
public:
    CWaveCounter(int nBase, int nAmp, int nCycle)
        : m_nBase(nBase), m_nAmp(nAmp), m_nCycle(nCycle), m_nCount(0) {}
    int GetAmp() const { return m_nAmp; }
    int GetCycle() const { return m_nCycle; }
    int Get()  const {
        return m_nBase + static_cast<int>(sin(6.2831853071795864769 * m_nCount / m_nCycle) * m_nAmp);
    }
    void Inc() { m_nCount++; }
private:
    int     m_nBase;
    int     m_nAmp;
    int     m_nCycle;
    int     m_nCount;
};

}
