#include "stdafx.h"
#include <math.h>
#include "Thread/XTime.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
#include "Timer.h"
#include "CursorMover.h"
namespace Themis {

CCursorMover::CCursorMover() : m_nFrameCount(-1), m_nWait(17), m_nAccel(8) {
    Themis_DebugBreak();
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&CCursorMover::OnThread, this));
}

CCursorMover::~CCursorMover() {
    {
        boost::mutex::scoped_lock lock(mutex);
        threadValid = false;
        condition.notify_all();
    }
    thread->Join();
}

CPoint CCursorMover::GetPos() const {
    CPoint pt;
    ::GetCursorPos(&pt);
    return pt;
}

void CCursorMover::SetPos(CPoint ptScreen) {
    while (IsMoving()) {
        ::Sleep(1); //Ȃ肢ȑ҂ǁACɂȂŁBB
    }
    boost::mutex::scoped_lock lock(mutex);
    // B
    ::GetCursorPos(&m_ptOrigin);
    m_ptNow = m_ptOrigin;
    m_szDistance = CSize(ptScreen - m_ptOrigin);
    m_nFrameCount = 0;  //̎_œĂ鎖ɂƂ
    // XbhĊJB
    condition.notify_all();
}

void CCursorMover::OnThread() {
/*
    v:xAx:ړsNZAt:o߃t[Aa:x(萔)Av0:x(萔)̂ƂA
        v = a * t + v0
        x = v * t = a * t^2 / 2 + v0 * t
    ŁAt = ta ̂Ƃ v = 0 ƂȂ v0 
        v0 = -a * ta
    ̂Ƃ x = xa ƂƁA
        xa = a * ta^2 / 2 + v0 * ta 
           = a * ta^2 / 2 - a * ta^2
           = -a * ta^2 / 2
    āAړsNZ xa ɑ΂鑍o߃t[ ta ́A
        ta = sqrt(2 * xa / (-a))
    ܂ACӃt[ t ɑ΂ ړsNZ x 
        x = a * t^2 / 2 + v0 * t  A
          = a * t * (t / 2 - ta)
          = (-a) * t * (ta - t / 2)

    ŁA
        xa = 540Ata = 60 ƂƁAa ̒l -0.3
        xa = 360Ata = 60 ƂƁAa ̒l -0.2
        xa = 180Ata = 60 ƂƁAa ̒l -0.1
        xa = 180Ata = 30 ƂƁAa ̒l -0.4
        xa = 180Ata = 15 ƂƁAa ̒l -1.6
    ȏAx̒Pʂ̓sNZAt̒Pʂ̓t[A
    a ̒l͂Ƃ肠-16{ĎRňɂƂB
*/
    Thread::SetPriority(Thread::ePriorityMax);
    Thread::SetName(L"CursorMoverThread");
    while (true) {
        {
            boost::mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            while (m_nFrameCount != 0) {
                m_nFrameCount = -1; //ĂȂƂ͕l
                condition.wait(lock);
                if (!threadValid) return;
            }
            Themis_DebugAssert(m_nFrameCount == 0);
        }

        double dDistance = sqrt(static_cast<double>(m_szDistance.cx * m_szDistance.cx + m_szDistance.cy * m_szDistance.cy));
        int nAccel = m_nAccel;
        double dMaxFrame = sqrt(2 * dDistance * 16 / nAccel); //ta = sqrt(2 * xa / (-a))
        int nMaxFrame = static_cast<int>(dMaxFrame) + 1;

        unsigned nTime = CTimer::Get();
        while (true) {
            if (nMaxFrame <= ++m_nFrameCount) {
                m_ptNow = CPoint(m_ptOrigin + m_szDistance);
                ::SetCursorPos(m_ptNow.x, m_ptNow.y);
                break;  //IB
            } else {
                double d = static_cast<double>(nAccel)/16 * m_nFrameCount * (dMaxFrame - m_nFrameCount / 2.0); //x = a * t * (t / 2 - ta)
                m_ptNow.x = m_ptOrigin.x + static_cast<int>(floor(m_szDistance.cx * d / dDistance + 0.5));
                m_ptNow.y = m_ptOrigin.y + static_cast<int>(floor(m_szDistance.cy * d / dDistance + 0.5));
                ::SetCursorPos(m_ptNow.x, m_ptNow.y);
            }

            nTime += m_nWait;
            {
                boost::mutex::scoped_lock lock(mutex);
                if (!threadValid) return;
                condition.timed_wait(lock, XTime::AfterMilliSeconds(nTime - CTimer::Get()));
                if (!threadValid) return;
            }

            CPoint ptNow = GetPos();
            if (ptNow != m_ptNow) {
                break;    // Ń}EXĂ炻ŃXgbv
            }
        }
    }
}

}
