#include "stdafx.h"
#include "Utility/StringUtility.h"
#include "ErrorHandler.h"
#include "Wnd.h"
#include "_config.h"
#include <signal.h>
namespace Themis {

namespace {
  #if Themis_ConfigErrorHandler
    static bool InitializeErrorHandler() {
        CErrorHandler::Register();
        return true;
    }
    volatile bool errorHandlerInitialized = InitializeErrorHandler();
  #else
    const bool errorHandlerInitialized = false;
  #endif
}

void CErrorHandler::OnAbort(int sig) {
    CWnd::RemoveQuitMsg();
    Themis_DebugWarning(L"ُI");
    ::MessageBox(NULL, "ُI", "G[", MB_OK | MB_ICONERROR | MB_TOPMOST);
    exit(1);
}

void CErrorHandler::OnFpe(int sig) {
    volatile double d = 0.0f;
    signal(SIGFPE, &OnFpe);
    Themis_DebugWarning(FPEException().what());
    throw FPEException();
}

void CErrorHandler::OnIll(int sig) {
    signal(SIGILL, &OnIll);
    Themis_DebugWarning(InvalidProgramException().what());
    throw InvalidProgramException();
}

void CErrorHandler::OnSegV(int sig) {
    signal(SIGSEGV, &OnSegV);
    Themis_DebugWarning(AccessViolationException().what());
    throw AccessViolationException();
}

void CErrorHandler::OnTerminate() {
/*
    * O@\[U[֐ (IuWFNg̃fXgN^܂) Ăяo̊֐ߊl̗Oc܂܁Aʂ̖ߊl̗Ô߂ɏI
    * oꂽÕnhO@\Ȃ
    * ǏIȃIuWFNgÓIȃIuWFNgƂđ݂ĂԂɁA̍\z܂͔j󂪁AOɂI
    * atexit() œo^ꂽ֐̎sOɂI
    * IyhȂ throw Ođo悤ƂAO͌ݏĂȂ
    * unexpected() ֐ȑOɈᔽ̂Ow肪ȂO𑗏oAstd::bad_exception ̗OwɊ܂܂ĂȂ
    * unexpected() ̃ftHgłĂяoĂ
    http://docs.sun.com/db/doc/805-7887/6j7dsdhf8?l=ja&a=view B
*/
    Themis_DebugWarning(L"terminate()Ă΂ꂽ");
    abort();
}

void CErrorHandler::OnUnexpected() {
/*
    OwɂȂOoɌĂяo炵B
*/
    set_unexpected(&OnUnexpected);
    Themis_DebugWarning(UnexpectedException().what());
    throw UnexpectedException();
}

void CErrorHandler::Register() {
    signal(SIGABRT, &CErrorHandler::OnAbort);
    signal(SIGFPE, &CErrorHandler::OnFpe);
    signal(SIGILL, &CErrorHandler::OnIll);
    signal(SIGSEGV, &CErrorHandler::OnSegV);
    set_terminate(&CErrorHandler::OnTerminate);
    set_unexpected(&CErrorHandler::OnUnexpected);
    Themis_DebugOutput(L"VOio^܂: ThreadID = " + ItoA(::GetCurrentThreadId()));
}

void CErrorHandler::RegisterOnUserThread() {
    if (errorHandlerInitialized) {
        Register();
    }
}

}
