#include "../../ReuseAllocator.h"
namespace Themis {

namespace {
    boost::mutex& GetAllocatorMutex() {
        static boost::mutex mutex;
        return mutex;
    }
    CReuseAllocator& GetAllocator() {
        static CReuseAllocator s_oAllocator;
        return s_oAllocator;
    }
    int& GetGarbageCount() {
        static int s_n = 0;
        return s_n;
    }
    inline void* MemoryPlaneAlloc(size_t n) {
        boost::mutex::scoped_lock lock(GetAllocatorMutex());
        return GetAllocator().Alloc(n);
    }
    inline void MemoryPlaneFree(void* p) {
        boost::mutex::scoped_lock lock(GetAllocatorMutex());
        GetAllocator().Free(p);
    }
}

CMemoryPlaneImpl::CMemoryPlaneImpl() : memory(NULL) {
}

CMemoryPlaneImpl::~CMemoryPlaneImpl() {
    Delete();
}

bool CMemoryPlaneImpl::IsCreated() const {
    return memory != NULL && 0 < size.cx && 0 < size.cy;
}

void CMemoryPlaneImpl::Create(const CSize& sz, int nBpp) {
    Themis_DebugAssert((nBpp % 8) == 0);
    try {
        size = sz;
     // size.cx = size.cx + 3 & ~3;
        pitch = size.cx * (nBpp / 8);
        pitch = (pitch + 3) & ~3;
        size_t nMemorySize = pitch * size.cy;
        if (memory != NULL) {
            MemoryPlaneFree(memory);
            memory = NULL;
        }
        memory = static_cast<uint8_t*>(MemoryPlaneAlloc(nMemorySize));
        // KxGarbage()B^C~OIɃC}C`邯ǁBB
        if (++GetGarbageCount() % 4 == 0) {
            GetAllocator().Garbage();
        }
    } catch (std::exception&) {
        if (memory != NULL) {
            MemoryPlaneFree(memory);
            memory = NULL;
        }
        throw RuntimeException(L"摜ێ̍쐬Ɏs܂");
    }
}

void CMemoryPlaneImpl::Delete() {
    if (memory != NULL) {
        MemoryPlaneFree(memory);
        memory = NULL;
    }
    pitch = 0;
    size = CSize();
}

CSize CMemoryPlaneImpl::GetSize() const {
    return size;
}

pair<void*, int> CMemoryPlaneImpl::Lock() {
    return make_pair(memory, pitch);
}

}
