#include "stdafx.h"
#include "SpiLoader.h"
#include "../ErrorHandler.h"
#include "../File.h"
#include "../Graphics.h"
#include "../Windows/PushErrorMode.h"
#include "../Utility/StringUtility.h"
#include "../String/Encoding.h"
namespace Themis {

namespace {
    // SPI}`XbhɑΉĂƂ͌Ȃ̂ŁAÔߔrƂ
    boost::mutex& GetSpiLoaderExclusion() {
        static boost::mutex mutex;
        return mutex;
    }
}

CSpiLoader::CSpiLoader(const std::tr1::shared_ptr<CSpi>& pSpi): m_pSpi(pSpi), m_hBits(NULL), m_hBmInfo(NULL), m_nAlphaFlag(0) {
}

CSpiLoader::~CSpiLoader() {
    try {
        if (m_oPictureInfo.hInfo != NULL) {
            ::LocalFree(m_oPictureInfo.hInfo);
        }
        if (m_hBmInfo != NULL) {
            ::LocalFree(m_hBmInfo);
        }
        if (m_hBits != NULL) {
            ::LocalFree(m_hBits);
        }
    } catch (...) {
        // Ȃ񂩂̕ӃNbV肷ꍇ邩BB
    }
}

bool CSpiLoader::Load(const std::tr1::shared_ptr<CFileData>& fileData, bool bRenderOnlyAtOnce) {
    boost::mutex::scoped_lock lock(GetSpiLoaderExclusion());
    PushErrorMode errorMode(SEM_FAILCRITICALERRORS | SEM_NOALIGNMENTFAULTEXCEPT |
        SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX);

    try {
        const vector<uint8_t>& fileMemory = fileData->GetMemory();
        if (fileMemory.size() < 2048) {
            // t@CȂ疖[߁B
            vector<uint8_t> tmp(4096); // SPI̎dlɂ2KbytemۂƏĂ邯ǔÔߑ傫߂ɁB
            std_copy(fileMemory.begin(), fileMemory.end(), tmp.begin());
            std::fill(tmp.begin() + fileMemory.size(), tmp.end(), 0);
            return InnerLoad(fileData->GetName(), &tmp[0], fileMemory.size());
        }
        return InnerLoad(fileData->GetName(), &fileMemory[0], fileMemory.size());
    } catch (...) {
        return false;
    }
}

namespace {
    static int WINAPI DummySpiCallback(int nNum,int nDenom,long lData) { return 0; }
}

bool CSpiLoader::InnerLoad(const CFileName& fileName, const void* pMemory, size_t nSize) {
    for (int i = 0; i < m_pSpi->GetCount(); i++) {
        if (!m_pSpi->IsImportFilter(i)) {
            continue;   // 摜pȂȂ玟
        }
        try {
            SPI::IsSupported fnIsSupported = m_pSpi->IsSupported(i);
            SPI::GetPictureInfo fnGetPictureInfo = m_pSpi->GetPictureInfo(i);
            SPI::GetPicture fnGetPicture = m_pSpi->GetPicture(i);
            if (fnIsSupported == NULL || fnGetPictureInfo == NULL || fnGetPicture == NULL) {
                Themis_DebugWarning(L"SPI̊֐擾Ɏs");
                continue;
            }
            boost::scoped_array<char> sz(ToMultiByteChars(fileName.Get()));

        BOOST_STATIC_ASSERT(sizeof(DWORD) == 4 && sizeof(void*) == 4);
        #pragma warning(push)
        #pragma warning(disable : 4311)
            if (0 == fnIsSupported(sz.get(), reinterpret_cast<DWORD>(pMemory))) {
                continue;   // vOCΉ
            }
        #pragma warning(pop)

            // 摜̓ǂݍ
            memset(&m_oPictureInfo, 0x00, sizeof(m_oPictureInfo));  //Ô߃[߁B
            m_hBmInfo = NULL; // ÔNULL
            m_hBits = NULL; // ÔNULL
            fnGetPictureInfo((LPSTR)(pMemory), (long)nSize, 1, &m_oPictureInfo);
            if (0 != fnGetPicture((LPSTR)(pMemory), (long)nSize, 1, &m_hBmInfo, &m_hBits,
                reinterpret_cast<FARPROC>(&DummySpiCallback), NULL)) {
                // s͉Ȃ
                memset(&m_oPictureInfo, 0x00, sizeof(m_oPictureInfo));  //Ô߃[߁B
                m_hBmInfo = NULL; // ÔNULL
                m_hBits = NULL; // ÔNULL
                continue;   // ܂ŗĎs邿Ƃ͔jt@CƂȋC邯ǈꉞցc
            }
        } catch (...) {
            // ǂݍݎs
            continue;
        }
        // 摜t@C`̎擾
        try {
            SPI::GetPluginInfo fnGetPluginInfo = m_pSpi->GetPluginInfo(i);
            Themis_DebugAssert(fnGetPluginInfo != NULL);
            m_strTypeName = string::Empty;
            for (int i = 0; ; i++) {
                char sz[0x100] = {0};
                if (0 == fnGetPluginInfo(i*2 + 3, sz, sizeof(sz))) {
                    break;
                }
                if (!string::IsNullOrEmpty(m_strTypeName)) {
                    m_strTypeName += L" / ";
                }
                m_strTypeName += Encoding::GetDefault().GetString(sz);
            }
        } catch (...) {
            m_strTypeName = L"Unknown SPI";
        }
        // ݒɂƂ
        m_nAlphaFlag = 0;
        return true;
    }
    return false;
}

CSize CSpiLoader::GetSize() const {
    return CSize(m_oPictureInfo.width, m_oPictureInfo.height);
}

void CSpiLoader::Render(CPlane& dst, const CPoint& pt, bool bRenderOnlyAtOnce) {
    BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock(m_hBmInfo);
    void* pBits = (void*)::LocalLock(m_hBits);
    try {
        m_nAlphaFlag = pBmpInfo->bmiHeader.biBitCount == 32 ? 2 : 1;
        dst.BltFromDIBits(*pBmpInfo, pBits, pt, CRect(CSize::GetLarge()));
    } catch (...) {
        try {
            ::LocalUnlock(m_hBmInfo);
            ::LocalUnlock(m_hBits);
        } catch (...) {
        }
        throw;
    }
    try {
        ::LocalUnlock(m_hBmInfo);
        ::LocalUnlock(m_hBits);
    } catch (...) {
    }
}

bool CSpiLoader::HasAlpha() const {
    Themis_DebugAssert(0 < m_nAlphaFlag);
    return m_nAlphaFlag == 2;
}

string CSpiLoader::GetTypeName() const {
    return string::IsNullOrEmpty(m_strTypeName) ? L"SPI" : m_strTypeName;
}

string CSpiLoader::GetTypeFileName() const {
    boost::mutex::scoped_lock lock(GetSpiLoaderExclusion());
    string strResult;
    try {
        for (int i = 0; i < m_pSpi->GetCount(); i++) {
            if (!m_pSpi->IsImportFilter(i)) {
                continue;   // 摜pȂȂ玟
            }
            try {
                SPI::GetPluginInfo gpi = m_pSpi->GetPluginInfo(i);
                Themis_DebugAssert(gpi != NULL);
                for (int i = 0; ; i++) {
                    char sz[0x80] = {0};
                    int n = gpi(i*2 + 2, sz, sizeof(sz) - 1);
                    if (n == 0) {
                        break;
                    }
                    sz[n] = L'\0';

                    if (!string::IsNullOrEmpty(strResult)) {
                        strResult += L";";
                    }
                    strResult += Encoding::GetDefault().GetString(sz);
                }
            } catch (std::exception&) {
                // ֐|C^̎擾sH
            }
        }
    } catch (...) {
        Themis_DebugWarning(L"Ȃ񂩗ONۂ");
    }
    return strResult;
}

std::tr1::shared_ptr<IImageLoader> CSpiLoaderFactory::CreateInstance() {
    return std::tr1::shared_ptr<IImageLoader>(new CSpiLoader(m_pPluginList));
}

void CSpiLoaderFactory::SetDir(const CFileName& oSpiDir) {
    if (m_pPluginList == NULL) {
        m_pPluginList.reset(new CSpi(oSpiDir));
    } else {
        m_pPluginList->SetDir(oSpiDir);
    }
}

}
