#include "stdafx.h"
#include "ImageList.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
#include "String/Encoding.h"
#include <commctrl.h>
#include <shellapi.h>
#pragma comment(lib, "comctl32.lib")
#pragma comment(lib, "shell32.lib")
namespace Themis {

CImageList::CImageList(const CSize& szImage, int nCapacity) {
    m_hImageList = ImageList_Create(szImage.cx, szImage.cy, ILC_COLOR32 | ILC_MASK, nCapacity, 0);
    ImageList_SetBkColor((HIMAGELIST)m_hImageList, ::GetSysColor(COLOR_WINDOW));
 // ImageList_SetBkColor((HIMAGELIST)m_hImageList, CLR_NONE);
}

CImageList::~CImageList() {
    if (m_hImageList != NULL) {
        ImageList_Destroy((HIMAGELIST)m_hImageList);
        m_hImageList = NULL;
    }
}

int CImageList::Add(HICON hIcon) {
    return ImageList_AddIcon((HIMAGELIST)m_hImageList, hIcon);
}

int CImageList::Add(HBITMAP hBmp, HBITMAP hBmpMask) {
    return ImageList_Add((HIMAGELIST)m_hImageList, hBmp, hBmpMask);
}

int CImageList::Add(HBITMAP hBmp, COLORREF rgbMask) {
    return ImageList_AddMasked((HIMAGELIST)m_hImageList, hBmp, rgbMask);
}

//------------------------------------------------------------------//

CIconImageList::CIconImageList(bool bSmall): m_bSmall(bSmall) {
    // B
    m_pImageList.reset(new CImageList(m_bSmall? CSize(16, 16): CSize(32, 32), 16));
}

CIconImageList::~CIconImageList() {
}

void CIconImageList::SetSize(int n) {
    m_pImageList.reset(new CImageList(m_bSmall? CSize(16, 16): CSize(32, 32), n));
}

int CIconImageList::Add(const CFileName& oFileName) {
    SHFILEINFO fi = {0};
    ::SHGetFileInfo(Encoding::GetDefault().GetBytes(oFileName.Get()).c_str(), 0, &fi, sizeof(fi), SHGFI_ICON | SHGFI_SYSICONINDEX | (m_bSmall? SHGFI_SMALLICON: 0));
    int nIndex = m_pImageList->Add(fi.hIcon);
    ::DestroyIcon(fi.hIcon);
    return nIndex;
}

//------------------------------------------------------------------//

CIconLoader::CIconLoader(IListener& rListener, bool bSmall) : m_rListener(rListener) {
    m_pIconImageList.reset(new CIconImageList(bSmall));
    // XbhJn
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&CIconLoader::OnThread, this));
}

CIconLoader::~CIconLoader() {
    {
        boost::mutex::scoped_lock lock(mutex);
        threadValid = false;
        condition.notify_all();
    }
    thread->Join();
}

/// ő̐ݒ(ǂݍ݂̓LZ)
void CIconLoader::SetSize(int n) {
    boost::mutex::scoped_lock lock(mutex);
    m_pIconImageList->SetSize(n);
    m_oLoadQueue.clear();
    m_bCancel = true;
}

/// ǂݍ݃X^bNɐς
void CIconLoader::Load(const CFileName& oFileName, int nParam) {
    boost::mutex::scoped_lock lock(mutex);
    SEntry s;
    s.oFileName = oFileName;
    s.nParam = nParam;
    m_oLoadQueue.push_back(s);
    condition.notify_all();
}

void CIconLoader::OnThread() {
    Thread::SetName(L"IconLoaderThread");
    while (true) {
        SEntry s;
        {
            boost::mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            while (m_oLoadQueue.empty()) {
                // Ȃҋ@
                condition.wait(lock);
                if (!threadValid) return;
            }
            // ǂݍݏ
            s = m_oLoadQueue.front();
            m_oLoadQueue.pop_front();
            m_bCancel = false;
        }
        // ǂݍ
        int nIndex = m_pIconImageList->Add(s.oFileName);
        // ǂݍ݌̃R[obN
        {
         // boost::mutex::scoped_lock lock(mutex);
         // OnIconLoaded()̎ɂĂ̓fbhbNꍇ蓾̂Ŏ~߂Ƃ
            if (!threadValid) return;
            if (!m_bCancel) {
                m_rListener.OnIconLoaded(s.oFileName, s.nParam, nIndex);
            }
        }
    }
}

}
