#pragma once
#include "Utility/Algorithm.h"
namespace Themis {

/**
 * `̈ɋ`zuWvZ̃C^[tF[X
 *
 * züScreenAzu`DocumentƌĂԁB
 */
class ILayout : private boost::noncopyable {
    CRect document, screen;     // `挳Ŝƕ`S
    mutable CRect src, dst;     // `挳ƕ`
    mutable bool modified;
public:
    ILayout() : modified(false) {}
    virtual ~ILayout() {}

    /// Document
    const CRect& GetDocument() const { return document; }
    /// Document
    void SetDocument(const CRect& rc) { document = rc; SetModified(); }
    /// Screen
    const CRect& GetScreen() const { return screen; }
    /// Screen
    void SetScreen(const CRect& rc) { screen = rc; SetModified(); }

    //@{ Update()ɂvZ
    /// `挳̋`(Document͂ݏoȂ`)
    const CRect& GetSrc() const { UpdateIfModified(); return src; }
    /// `̋`(Screen͂ݏoȂ`)
    const CRect& GetDst() const { UpdateIfModified(); return dst; }
    //@}

    /// DocumentŜɑΉ]TCY
    CSize GetVirtualDstSize() const {
        return DocumentToScreen(GetDocument().GetSize());
    }
    /// GetVirtualDstSize().ToRect()DocumentƌȂƂSrc̍W
    CRect GetVirtualDstSrc() const {
        CRect src = GetSrc(), dst = GetDst();
        CSize srcSize = src.GetSize(), dstSize = dst.GetSize();
        if (0 < srcSize.cx && 0 < srcSize.cy) {
            CPoint scrollPos = src.GetPoint() - GetDocument().GetPoint();
            CPoint dstPos(MulDiv45(scrollPos.x, dstSize.cx, srcSize.cx),
                          MulDiv45(scrollPos.y, dstSize.cy, srcSize.cy));
            return CRect(dstPos, dstSize);
        } else {
            return CRect(CPoint(), dstSize);
        }
    }
    /// XN[̎擾B0ȉȂXN[B
    CSize GetScrollRanges() const { return GetDocument().GetSize() - GetSrc().GetSize(); }
    /// XN[̗Lc܂Ƃ߂ĔB0x01̃rbgĂ牡XN[LB0x02͏cB
    int GetScrollType() const {
        CSize ranges = GetScrollRanges();
        return (ranges.cx <= 0 ? 0x00 : 0x01) | (ranges.cy <= 0 ? 0x00 : 0x02);
    }
    /// XN[ʒuB(0,0) ` GetScrollRanges() ̒lBUpdate()Ă΂ƃZbgB
    CPoint GetScrollPos() const {
        CPoint scrollPos = GetSrc().GetPoint() - GetDocument().GetPoint();
        return ClipScrollPos(scrollPos);
    }
    /// XN[B
    void SetScrollPos(const CPoint& scrollPos) {
        CPoint pt = ClipScrollPos(scrollPos);
        // GetScrollPos()̋tZ
        src.SetPoint(CPoint(pt + GetDocument().GetPoint()));
    }
    /// XN[ʒũNbsO
    CPoint ClipScrollPos(const CPoint& scrollPos) const {
        CSize ranges = GetScrollRanges();
        return CPoint(Max(Min(scrollPos.x, ranges.cx), 0L),
                      Max(Min(scrollPos.y, ranges.cy), 0L));
    }

    //@{ Wϊȏ
    /// Screenł̍WDocumentł̍W֕ϊ
    CPoint ScreenToDocument(const CPoint& pt) const {
        CRect src = GetSrc(), dst = GetDst();
        CSize srcSize = src.GetSize(), dstSize = dst.GetSize();
        if (0 < dstSize.cx && 0 < dstSize.cy) {
            return CPoint(MulDiv45((pt.x - dst.left), srcSize.cx, dstSize.cx) + src.left,
                          MulDiv45((pt.y - dst.top),  srcSize.cy, dstSize.cy) + src.top);
        } else {
            return CPoint();
        }
    }
    /// Screenł̑傫Documentł̑傫֕ϊ
    CSize ScreenToDocument(const CSize& sz) const {
        CRect src = GetSrc(), dst = GetDst();
        CSize srcSize = src.GetSize(), dstSize = dst.GetSize();
        if (0 < dstSize.cx && 0 < dstSize.cy) {
            return CSize(MulDiv45(sz.cx, srcSize.cx, dstSize.cx),
                         MulDiv45(sz.cy, srcSize.cy, dstSize.cy));
        } else {
            return CSize();
        }
    }
    /// Screenł̋`Documentł̋`֕ϊ
    CRect ScreenToDocument(const CRect& rc) const {
        return CRect(ScreenToDocument(rc.GetPoint()), ScreenToDocument(rc.GetSize()));
    }
    /// Documentł̍WScreenł̍W֕ϊ
    CPoint DocumentToScreen(const CPoint& pt) const {
        CRect src = GetSrc(), dst = GetDst();
        CSize srcSize = src.GetSize(), dstSize = dst.GetSize();
        if (0 < srcSize.cx && 0 < srcSize.cy) {
            return CPoint(MulDiv45((pt.x - src.left), srcSize.cx, srcSize.cx) + dst.left,
                          MulDiv45((pt.y - src.top),  srcSize.cy, srcSize.cy) + dst.top);
        } else {
            return CPoint();
        }
    }
    /// Documentł̑傫Screenł̑傫֕ϊ
    CSize DocumentToScreen(const CSize& sz) const {
        CRect src = GetSrc(), dst = GetDst();
        CSize srcSize = src.GetSize(), dstSize = dst.GetSize();
        if (0 < srcSize.cx && 0 < srcSize.cy) {
            return CSize(MulDiv45(sz.cx, dstSize.cx, srcSize.cx),
                         MulDiv45(sz.cy, dstSize.cy, srcSize.cy));
        } else {
            return CSize();
        }
    }
    /// Documentł̋`Screenł̋`֕ϊ
    CRect DocumentToScreen(const CRect& rc) const {
        return CRect(DocumentToScreen(rc.GetPoint()), DocumentToScreen(rc.GetSize()));
    }
    //@}

    /// g嗦̎擾
    double GetRateH() const {
        CRect src = GetSrc(), dst = GetDst();
        return src.GetWidth() <= 0 ? 0.0 : static_cast<double>(dst.GetWidth()) / src.GetWidth();
    }
    /// g嗦̎擾
    double GetRateV() const {
        CRect src = GetSrc(), dst = GetDst();
        return src.GetHeight() <= 0 ? 0.0 : static_cast<double>(dst.GetHeight()) / src.GetHeight();
    }
    /// {Ȃtrue
    bool IsActualSize() const { return GetDst().GetSize() == GetSrc().GetSize(); }

    /// ̍쐬
    virtual std::tr1::shared_ptr<ILayout> Clone() const = 0;

protected:
    /// CAEgČvZKvȂĂԁB
    void SetModified() { modified = true; }

    /// ľܓȃwp֐
    static inline int Div45(int x, int y) { return (x + y / 2) / y; }
    /// ľܓȃwp֐
    static inline int MulDiv45(int x, int y, int z) { return Div45(x * y, z); }

    /// SetModified()ĂSrcADstAVirtualDstČvZ
    void UpdateIfModified() const {
        if (modified) {
            Update();
            modified = false;
            // ̕ӂɈĂ̂Update()rȂ񂩂̃oOH
            Themis_DebugAssert(!src.IsValid() || GetDocument().IsIn(src));
            Themis_DebugAssert(!dst.IsValid() || GetScreen().IsIn(dst));
        }
    }

private:
    /// Screen/Document̒lDst/Src/VirtualDstZoBScrollPos̓ZbĝŒ
    void Update() const;

    /// DocumentɑΉ]W̎Zo
    virtual CRect CalculateVirtualDst() const = 0;
};


/**
 * GȏsȂCAEgB
 *
 * [ɃNbsOs̏ꍇɎgB
 * Clipper͂ƎgÂ炢Ƃ낪̂ŁBB
 */
class SimpleLayout : public ILayout {
    CRect documentViewArea;
public:
    /// Document̕\GA
    const CRect& GetDocumentViewArea() const { return documentViewArea; }
    /// Document̕\GA
    void SetDocumentViewArea(const CRect& rc) { documentViewArea = rc; SetModified(); }
    /// DocumentŜ\GA
    void SetDocumentViewAreaAll() { SetDocumentViewArea(GetDocument()); }

    /// ̍쐬
    std::tr1::shared_ptr<ILayout> Clone() const;

private:
    /// DocumentɑΉ]W̎Zo
    virtual CRect CalculateVirtualDst() const;
};

/**
 * AutoStretchingLayoutƂ̃x[XNXB
 */
class StretchingLayoutBase : public ILayout {
    int alignV, alignH;
public:
    StretchingLayoutBase() : alignV(1), alignH(1) {}

    /// Update̐zu@B0:񂹁A1:񂹁A2:
    int GetAlignV() const { return alignV; }
    /// Update̐zu@B0:񂹁A1:񂹁A2:
    void SetAlignV(int n) { alignV = n; SetModified(); }

    /// Update̐zu@B0:񂹁A1:񂹁A2:E
    int GetAlignH() const { return alignH; }
    /// Update̐zu@B0:񂹁A1:񂹁A2:E
    void SetAlignH(int n) { alignH = n; SetModified(); }

private:
    /// DocumentɑΉ]W̎Zo
    virtual CRect CalculateVirtualDst() const;
    /// DocumentɑΉ]TCY̎Zo
    virtual CSize CalculateVirtualDstSize() const = 0;
};

/**
 * Screenɍ킹gktzuB
 *
 * bStickOut = false ŁA͂ݏoȂő̎kA
 * bStickOut = true ŁAScreenŜɕ\Œ̎kpB
 */
class AutoStretchingLayout : public StretchingLayoutBase {
    bool m_bReduction;
    bool m_bMagnification;
    bool m_bStickOut;
    mutable bool stretchToHorz;
public:
    AutoStretchingLayout() : m_bReduction(false), m_bMagnification(false), m_bStickOut(false) {}

    /// k̂ǂ
    bool IsReduction() const { return m_bReduction; }
    /// k̂ǂ
    void SetReduction(bool b) { m_bReduction = b; SetModified(); }

    /// g傷̂ǂ
    bool IsMagnification() const { return m_bMagnification; }
    /// g傷̂ǂ
    void SetMagnification(bool b) { m_bMagnification = b; SetModified(); }

    /// ͂ݏôǂ
    bool IsStickOut() const { return m_bStickOut; }
    /// ͂ݏôǂ
    void SetStickOut(bool b) { m_bStickOut = b; SetModified(); }

    /// ɍ킹ĊgEkȂtrue
    bool IsStretchToHorz() const { UpdateIfModified(); return stretchToHorz; }

    /// ̍쐬
    std::tr1::shared_ptr<ILayout> Clone() const;

private:
    /// DocumentɑΉ]TCY̎Zo
    virtual CSize CalculateVirtualDstSize() const;
};

/**
 * {ŒzuB
 */
class FixedStretchingLayout : public StretchingLayoutBase {
    double rate;
public:
    FixedStretchingLayout() : rate(1.0) {}

    /// g嗦B1.0œ{
    double GetRate() const { return rate; }
    /// g嗦B1.0œ{
    void SetRate(double b) { rate = b; SetModified(); }

    /// ̍쐬
    std::tr1::shared_ptr<ILayout> Clone() const;

private:
    /// DocumentɑΉ]TCY̎Zo
    virtual CSize CalculateVirtualDstSize() const;
};

}
