#include "stdafx.h"
#include "TabCtrl.h"
#include "String/Encoding.h"
#include <commctrl.h>
#pragma comment(lib, "comctl32.lib")
namespace Themis {

CTabPage::CTabPage(int nResourceId) : CResourceDialogBase(nResourceId), m_pParentTabCtrl(NULL) {
}

CTabPage::~CTabPage() {
}

void CTabPage::CreatePage(Themis::IMsgHooker* pHooker, HWND hParent) {
    GetDialog().GetHook().Add(pHooker);
    OpenModeless(hParent, false);
    GetDialog().SetStyle(GetDialog().GetStyle() & ~WS_CAPTION);
    GetDialog().SetEnableThemeBg(true, true);
}

void CTabPage::ClosePage() {
    GetDialog().Close();
}

void CTabPage::ShowPage() {
    GetDialog().Show();
}

void CTabPage::HidePage() {
    GetDialog().Hide();
}

CSize CTabPage::GetPageSize() {
    return GetDialog().GetSize();
}

void CTabPage::SetPagePos(const CPoint& pt) {
    GetDialog().SetPos(pt);
}

bool CTabPage::HookMsg(CMsgHookPrms& rPrms) {
    if (rPrms.GetMsg() == WM_CLOSE) {
        GetDialog().Delete();
        return true;
    }
    return false;
}

/// ԂƂModifiedMarkt
void CTabPage::SetTabModifiedMark(bool bModified, const string& strSuffix) {
    Themis_DebugAssertIf(m_pParentTabCtrl != NULL) {
        int nPage = m_pParentTabCtrl->FindPage(*this);
        Themis_DebugAssertIf(0 <= nPage) {
            m_pParentTabCtrl->SetTextModifiedMark(nPage, bModified, strSuffix);
        }
    }
}

//------------------------------------------------------------------//

CTabCtrl::CTabCtrl(CWindow& rParent, int nCtrlId)
: CCtrl(rParent.GetCtrl(nCtrlId)), m_pOwner(NULL), m_nNowPage(0) {
    rParent.GetHook().Add(this);
    m_hWndParent = rParent.GetHandle();
 #if 0
    SetStyle(GetStyle() | WS_CLIPSIBLINGS);
 #endif
}

CTabCtrl::CTabCtrl(IOwner& rOwner, CWindow& rParent, int nCtrlId)
: CCtrl(rParent.GetCtrl(nCtrlId)), m_pOwner(&rOwner), m_nNowPage(0) {
    rParent.GetHook().Add(this);
    m_hWndParent = rParent.GetHandle();
}

CTabCtrl::~CTabCtrl() {
    for (size_t i = 0, n = m_vctPages.size(); i < n; i++) {
        m_vctPages[i]->ClosePage();
    }
}

void CTabCtrl::SetPage(int n) {
    HidePage(m_nNowPage);
    ShowPage(n);
    TabCtrl_SetCurSel(GetHandle(), n);
    m_nNowPage = n;
}

/// y[WSč폜
void CTabCtrl::ClearPages() {
    TabCtrl_DeleteAllItems(GetHandle());
}

/// wy[W폜
void CTabCtrl::ClearPage(int n) {
    TabCtrl_DeleteItem(GetHandle(), n);
}

void CTabCtrl::InsertPage(int n, const string& caption, const std::tr1::shared_ptr<CTabPage>& pPage) {
    if (GetPageCount() <= n) {
        m_vctPages.resize(n + 1);
    }
    m_vctPages[n] = pPage;
    m_vctPages[n]->m_pParentTabCtrl = this;
 #if 1
    // _CAO̎qɂāA^uƂ͌ZɂB̂ł܂ŝ񂪁cB
    m_vctPages[n]->CreatePage(this, m_hWndParent);
 #elif 0
    // ċAIɐeHĂ݂
    HWND hWnd = m_hWndParent;
    while (true) {
        HWND hWndTmp = ::GetParent(hWnd);
        if (hWndTmp == NULL) {
            m_vctPages[n]->CreatePage(this, hWnd);
            break;
        }
        hWnd = hWndTmp;
    }
 #else
    // ^u̎qɂĂ݂
    m_vctPages[n]->CreatePage(this, GetHandle());
 #endif
    if (n == m_nNowPage) {
        ShowPage(n);
    } else {
        HidePage(n);
    }
    std::string strCaption = Encoding::GetDefault().GetBytes(caption);
    TCITEM ti = {0};
    ti.mask = TCIF_TEXT;
    ti.pszText = const_cast<char*>(strCaption.c_str());
    ti.iImage = -1;
    TabCtrl_InsertItem(GetHandle(), n, &ti);
}

//@{ y[W̃^Cg擾Eݒ
string CTabCtrl::GetText(int n) const {
    char szBuf[0x100] = {0};
    TCITEM ti = {0};
    ti.mask = TCIF_TEXT;
    ti.pszText = szBuf;
    ti.cchTextMax = sizeof(szBuf) - 1; // -1vC邯ǔÔ
    ti.iImage = -1;
    if (TabCtrl_GetItem(GetHandle(), n, &ti) != FALSE) {
        return Encoding::GetDefault().GetString(ti.pszText);
    }
    return string::Empty;
}

void CTabCtrl::SetText(int n, const string& caption) {
    std::string strCaption = Encoding::GetDefault().GetBytes(caption);
    TCITEM ti = {0};
    ti.mask = TCIF_TEXT;
    ti.pszText = const_cast<char*>(strCaption.c_str());
    ti.iImage = -1;
    BOOL bRet = TabCtrl_SetItem(GetHandle(), n, &ti);
    Themis_DebugAssert(bRet != FALSE);
}

/// bModifiedtrueȂstrSuffix(L" *")t
void CTabCtrl::SetTextModifiedMark(int n, bool bModified, const string& strSuffix) {
    string strCaption = GetText(n);
    bool bCaptionModified = strCaption.EndsWith(strSuffix); // ɕtĂ̂ǂ
    if (bModified != bCaptionModified) {
        if (bModified) { // t
            SetText(n, strCaption + strSuffix);
        } else { // O
            int t = static_cast<int>(strCaption.GetLength() - strSuffix.GetLength());
            Themis_DebugAssertIf(0 <= t) {
                SetText(n, strCaption.Substring(0, t));
            }
        }
    }
}
//@}

/// SẴy[W܂ŏTCY̎擾
CSize CTabCtrl::GetMinSize() {
    Themis_DebugAssert(!m_vctPages.empty());
    CSize sz = m_vctPages[0]->GetPageSize();
    for (size_t i = 0, n = m_vctPages.size(); i < n; i++) {
        CSize sz2 = m_vctPages[0]->GetPageSize();
        if (sz2.cx < sz.cx) sz.cx = sz2.cx;
        if (sz2.cy < sz.cy) sz.cy = sz2.cy;
    }
    CRect rcTab;
    ::GetClientRect(::GetParent(GetHandle()), &rcTab);
    TabCtrl_AdjustRect(GetHandle(), TRUE, &rcTab);
    // rcTabɂ̓^u̕\̈悪ZbgĂ͂BB
    return GetSize() - rcTab.GetSize() + sz;
}

/// w^uy[W̃y[W擾B-1B
int CTabCtrl::FindPage(CTabPage& rTabPage) {
    for (size_t i = 0, n = m_vctPages.size(); i < n; i++) {
        if (m_vctPages[i].get() == &rTabPage) {
            return static_cast<int>(i);
        }
    }
    return -1;
}

bool CTabCtrl::HookMsg(CMsgHookPrms& rPrms) {
    if (rPrms.GetWnd() == m_hWndParent) { // ^uRg[̐eEBhEȏꍇ
        if (rPrms.GetMsg() == WM_NOTIFY) {
            NMHDR* hdr = reinterpret_cast<NMHDR*>(rPrms.GetLPrm());
            if (hdr->hwndFrom == GetHandle() && hdr->code == TCN_SELCHANGE) {
                int nOldPage = m_nNowPage;
                m_nNowPage = TabCtrl_GetCurSel(GetHandle());
                HidePage(nOldPage);
                ShowPage(m_nNowPage);
                return true;
            }
        }
    } else if (m_pOwner != NULL) { // ^uy[Wȏꍇ
        for (size_t i = 0, n = m_vctPages.size(); i < n; i++) {
            if (rPrms.GetWnd() == m_vctPages[i]->GetDialog().GetHandle()) {
                return m_pOwner->HookMsgTabPages(*this, int(i), rPrms);
            }
        }
    }
    return false;
}

void CTabCtrl::ShowPage(int n) {
    if (0 <= n && n < GetPageCount()) {
        CRect rcTab = GetRect();
        TabCtrl_AdjustRect(GetHandle(), FALSE, &rcTab);
      #if 0
        m_vctPages[n]->SetPagePos(rcTab.GetPoint());
      #else
        m_vctPages[n]->SetPagePos(rcTab.GetCenterOf(m_vctPages[n]->GetPageSize()));
      #endif
        m_vctPages[n]->ShowPage();
        if (m_pOwner != NULL) {
            m_pOwner->OnShowTabPage(*this, n);
        }
    }
}

void CTabCtrl::HidePage(int n) {
    if (0 <= n && n < GetPageCount()) {
        m_vctPages[n]->HidePage();
    }
}

}
