#include "stdafx.h"
#include "TempFile.h"
#include "File.h"
#include "FileBase.h"
#include "Windows/System.h"
#include "String/Encoding.h"
namespace Themis {

CTempFile::CTempFile(const string& strPrefix, const string& strSuffix) : m_oBaseDir(CSystem::GetTempDir()), m_bOwner(false) {
    m_oFileName = CFileName(strPrefix + string::Format(L"%x@%x.", this, ::GetCurrentProcessId()) + strSuffix, m_oBaseDir);
    m_bOwner = true;
}

CTempFile::CTempFile(const CFileName& oBaseDir, const string& strPrefix, const string& strSuffix) : m_oBaseDir(oBaseDir), m_bOwner(false) {
    m_oFileName = CFileName(strPrefix + string::Format(L"%x@%x.", this, ::GetCurrentProcessId()) + strSuffix, m_oBaseDir);
    m_bOwner = true;
}

CTempFile::~CTempFile() {
    Delete();
}

void CTempFile::Create(const vector<uint8_t>& vct) {
    if (CFile(m_oFileName).Exists()) {
        throw IOException(L"e|t@C '" + m_oFileName.Get() + L"' ̍쐬Ɏs܂");
    }
    CStdFile oFile(m_oFileName, L"wb");
    if (!vct.empty()) {
        oFile.Write(&vct[0], vct.size());
        oFile.Flush();
    }
}

void CTempFile::Create(const string& str) {
    if (CFile(m_oFileName).Exists()) {
        throw IOException(L"e|t@C '" + m_oFileName.Get() + L"' ̍쐬Ɏs܂");
    }
    std::string mb = Encoding::GetDefault().GetBytes(str);
    CStdFile oFile(m_oFileName, L"wb");
    oFile.Write(mb.c_str(), mb.length());
    oFile.Flush();
}

void CTempFile::CreateByCopy(const CFileName& oFileName) {
    if (CFile(m_oFileName).Exists()) {
        throw IOException(L"e|t@C '" + m_oFileName.Get() + L"' ̍쐬Ɏs܂");
    }
    CFile(oFileName).Copy(m_oFileName);
}

void CTempFile::CreateByMove(const CFileName& oFileName) {
    if (CFile(m_oFileName).Exists()) {
        throw IOException(L"e|t@C '" + m_oFileName.Get() + L"' ̍쐬Ɏs܂");
    }
    CFile(oFileName).Move(m_oFileName);
}

void CTempFile::Delete() {
    if (m_bOwner) {
        CFile oFile(m_oFileName);
        if (oFile.Exists()) {
            oFile.Delete();
        }
        m_bOwner = false;
    }
}

void CTempFile::Release() {
    m_bOwner = false;
}

/// oDstFileNameɈړRelease();
void CTempFile::MoveAndRelease(const CFileName& oDstFileName) {
    if (m_bOwner) {
        CFile oFile(m_oFileName);
        if (oFile.Exists()) {
            oFile.Move(oDstFileName);
        }
        m_bOwner = false;
    }
}

//////////////////////////////////////////////////////////////////////

CTempDir::CTempDir(const string& strPrefix): m_oBaseDir(CSystem::GetTempDir()) {
    m_oDirName = CFileName(strPrefix + string::Format(L"%x@%x\\", this, ::GetCurrentProcessId()), m_oBaseDir);
}

CTempDir::CTempDir(const CFileName& oBaseDir, const string& strPrefix): m_oBaseDir(oBaseDir) {
    m_oDirName = CFileName(strPrefix + string::Format(L"%x@%x\\", this, ::GetCurrentProcessId()), m_oBaseDir);
}

CTempDir::~CTempDir() {
    try {
        Delete();
    } catch (std::exception&) {
        // ߂ǂ̂ŖَEBc_ȂDelete()蓮ŌĂԕŁB(L`)
    }
}

void CTempDir::Create() {
    CFile(m_oDirName).CreateDir();
}

void CTempDir::Delete() {
    if (!m_oDirName.IsEmpty()) {
        CFile oDir(m_oDirName);
        if (oDir.Exists()) {
            oDir.Delete(true);
        }
        m_oDirName = CFileName();
    }
}

void CTempDir::Release() {
    m_oDirName = CFileName();
}

}
