#include "stdafx.h"
#include "Window.h"
#include "Windows/System.h"
#include "Windows/WinAPI.h"
#include "Thread/XTime.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
#include "String/Encoding.h"
#include "tpl/set_vector.h"
#include "tpl/map_vector.h"
#include <commctrl.h>
#pragma comment(lib, "comctl32.lib")
namespace Themis {

namespace Detail {
    struct SWindowThreadData {
        volatile bool threadValid;
        std::tr1::shared_ptr<WorkerThreadController> thread;
        boost::mutex mutex;
        boost::condition condition;

        volatile bool bWaitMsg;
        volatile bool bMsgLoopValid;
        SWindowThreadData() : bWaitMsg(false), bMsgLoopValid(false) {
            threadValid = true;
        }
    };
    struct SWindowCreateParam {
        typedef void(CWindow::* FCreate)(const IWindowParameter&);
        typedef const IWindowParameter  Parameter;
        FCreate     fnCreate;
        Parameter*  pParameter;
        string      strErrorMsg;
        SWindowCreateParam(FCreate f, Parameter* p): fnCreate(f), pParameter(p) {}
    };
}

CWindowParameter::CWindowParameter() {
    dwStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_BORDER | WS_MINIMIZEBOX | WS_CLIPCHILDREN;   //WS_VISIBLE͖ŁB
    dwExStyle = WS_EX_APPWINDOW;

    hMenu = NULL;
    hParent = NULL;
}

CWindow::CWindow() : m_pMsgHook(new CMsgHook), m_fnWndProc(NULL),
m_hAccel(NULL), m_nCursorState(0), result(DialogResult::None) {
    m_pMsgHook->Add(this);
    SetDefaultExceptionHandler();
}

CWindow::~CWindow() {
    Themis_DebugAssert(!Exists());
    //RDelete()Ă΂ɑIuWFNg폜ƂɈB
    //FXnȂ\̂ŁAWM_CLOSEƂ炵ĂԂׂB
    if (m_pThreadData.get() != NULL) {
        if (m_pThreadData->bMsgLoopValid) {
         // CloseForce();
            {
                boost::mutex::scoped_lock lock(m_pThreadData->mutex);
                m_pThreadData->threadValid = false;
                m_pThreadData->condition.notify_all();
            }
            m_pThreadData->thread->Join();
            PostMsg(WM_NULL, 0, 0); // ̂Ăꍇ̂ŁBB
        } else {
            //Ô߃XbhI܂ő҂Ƃ
            m_pThreadData->thread->Join();
        }
    }
}

void CWindow::Create(const CWindowParameter& rParameter) {
    Themis_DebugAssert(!Exists());
    Themis_DebugAssert(m_pThreadData.get() == NULL || !m_pThreadData->bMsgLoopValid);
    Detail::SWindowCreateParam cp(&CWindow::InnerCreate, &rParameter);
    Initialize(cp);
    if (!string::IsNullOrEmpty(cp.strErrorMsg)) {
        throw RuntimeException(cp.strErrorMsg);
    }
}

void CWindow::Create(const CDialogParameter& rParameter) {
    Themis_DebugAssert(!Exists());
    Themis_DebugAssert(m_pThreadData.get() == NULL || !m_pThreadData->bMsgLoopValid);
    Detail::SWindowCreateParam cp(&CWindow::InnerDialogCreate, &rParameter);
    Initialize(cp);
    if (!string::IsNullOrEmpty(cp.strErrorMsg)) {
        throw RuntimeException(cp.strErrorMsg);
    }
}

int CWindow::Create(const CModalDialogParameter& rParameter) {
    Themis_DebugAssert(!Exists());
    Themis_DebugAssert(m_pThreadData.get() == NULL || !m_pThreadData->bMsgLoopValid);
    m_eMode = EModeDialogModal;
    //_CAO쐬[
    INT_PTR nRet = NativeString::IsUnicode() ?
        ::DialogBoxParamW(::GetModuleHandleW(NULL), MAKEINTRESOURCEW(rParameter.GetId()),
        rParameter.GetParent(), &CWindow::StartDialogProc, reinterpret_cast<LPARAM>(this)) :
        ::DialogBoxParamA(::GetModuleHandleA(NULL), MAKEINTRESOURCEA(rParameter.GetId()),
        rParameter.GetParent(), &CWindow::StartDialogProc, reinterpret_cast<LPARAM>(this));
 // if ((nRet == 0 && ::GetLastError() != 0) || nRet == -1) {
    if (nRet == -1) {
        throw Win32Exception(L"_CAO̍쐬Ɏs");
    }
    Themis_DebugAssert(nRet == GetResult());
    // ̂cĂ̂ňꉞBB
    CWnd::RemoveQuitMsg();
    return static_cast<int>(nRet);
}

void CWindow::CreateWithThread(const CWindowParameter& rParameter) {
    Themis_DebugAssert(!Exists());
    Themis_DebugAssert(m_pThreadData.get() == NULL || !m_pThreadData->bMsgLoopValid);
    Detail::SWindowCreateParam cp(&CWindow::InnerCreate, &rParameter);
    m_pThreadData.reset(new Detail::SWindowThreadData);
    boost::mutex::scoped_lock lock(m_pThreadData->mutex);
    m_pThreadData->thread = ThreadPool::CreateWorkerThread();
    m_pThreadData->thread->SetWork(std::tr1::bind(&CWindow::OnThread, this, &cp));
    if (rParameter.GetParent() == NULL) {
        // 쐬܂ő҂
        m_pThreadData->condition.wait(lock);
    } else {
        // eEBhEȂŃnOĈŏA쐬܂ő҂
        boost::xtime sleepTime = XTime::AfterMilliSeconds(10);
        while (!m_pThreadData->condition.timed_wait(lock, sleepTime)) {
            if (!MsgPumping()) {
                throw RuntimeException(L"EBhE̍쐬Ɏs܂");
            }
            XTime::AddMilliSeconds(sleepTime, 10);
        }
    }
    if (!string::IsNullOrEmpty(cp.strErrorMsg)) {
        throw RuntimeException(cp.strErrorMsg);
    }
}

void CWindow::CreateWithThread(const CDialogParameter& rParameter) {
    Themis_DebugAssert(!Exists());
    Themis_DebugAssert(m_pThreadData.get() == NULL || !m_pThreadData->bMsgLoopValid);
    Detail::SWindowCreateParam cp(&CWindow::InnerDialogCreate, &rParameter);
    m_pThreadData.reset(new Detail::SWindowThreadData);
    boost::mutex::scoped_lock lock(m_pThreadData->mutex);
    m_pThreadData->thread = ThreadPool::CreateWorkerThread();
    m_pThreadData->thread->SetWork(std::tr1::bind(&CWindow::OnThread, this, &cp));
    if (rParameter.GetParent() == NULL) {
        // 쐬܂ő҂
        m_pThreadData->condition.wait(lock);
    } else {
        // eEBhEȂŃnOĈŏA쐬܂ő҂
        boost::xtime sleepTime = XTime::AfterMilliSeconds(10);
        while (!m_pThreadData->condition.timed_wait(lock, sleepTime)) {
            if (!MsgPumping()) {
                throw RuntimeException(L"_CAO̍쐬Ɏs܂");
            }
            XTime::AddMilliSeconds(sleepTime, 10);
        }
    }
    if (!string::IsNullOrEmpty(cp.strErrorMsg)) {
        throw RuntimeException(cp.strErrorMsg);
    }
}

namespace {
    template<class WNDCLASSEX, class STR> static inline
    void FillWndClassEx(WNDCLASSEX& wcex, const CWindowParameter& rParameter,
        HINSTANCE hInstance, const STR* classNameStr, WNDPROC wndProc) {
     // memset(&wcex, 0x00, sizeof(WNDCLASSEX));
        wcex.cbSize = sizeof(WNDCLASSEX);
        wcex.style = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        //  CS_SAVEBITS     OSɍĕ`悳
        //  CS_NOCLOSE      郁j[gps\ɂ
        wcex.hInstance = hInstance;
        wcex.hbrBackground = NULL;  //reinterpret_cast<HBRUSH>(COLOR_3DFACE + 1);
        wcex.hIcon = ::LoadIcon(NULL, IDI_APPLICATION);
        wcex.hCursor = ::LoadCursor(NULL, IDC_ARROW);   //Ăǁ[ׂȂ񂾂납BB
        wcex.lpszClassName = classNameStr;
        wcex.lpfnWndProc = wndProc;
    }
}
void CWindow::InnerCreate(const IWindowParameter& r) {
    const CWindowParameter& rParameter = static_cast<const CWindowParameter&>(r);
    m_eMode = EModeWindow;

    bool bRegister;
    string className;
    if (string::IsNullOrEmpty(rParameter.GetClassName())) {
        bRegister = true;
        className = string::Format(L"<%x:%p:%x>", ::GetCurrentProcessId(), this, rand());
        m_strClassName = className;
    } else {
        bRegister = false;
        className= rParameter.GetClassName();
        m_strClassName = string::Empty;
    }

    HWND hWnd;
    NativeString classNameStr = className;
    if (NativeString::IsUnicode()) {
        WNDCLASSEXW wc = {0};
        //NX̓o^
        if (bRegister) {
            FillWndClassEx(wc, rParameter, ::GetModuleHandleW(NULL), classNameStr.GetW(), &CWindow::StartWindowProc);
            if (::RegisterClassExW(&wc) == 0) {
                throw Win32Exception(L"EBhENX̓o^Ɏs");
            }
        }
        m_fnWndProc = &::DefWindowProcW;
        //EBhE̍쐬
        hWnd = ::CreateWindowExW(rParameter.GetExStyle(), wc.lpszClassName, NULL, rParameter.GetStyle(),
            CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, rParameter.GetParent(), rParameter.GetMenu(),
            wc.hInstance, this);
    } else {
        WNDCLASSEXA wc = {0};
        //NX̓o^
        if (bRegister) {
            FillWndClassEx(wc, rParameter, ::GetModuleHandleA(NULL), classNameStr.GetA(), &CWindow::StartWindowProc);
            if (::RegisterClassExA(&wc) == 0) {
                throw Win32Exception(L"EBhENX̓o^Ɏs");
            }
        }
        m_fnWndProc = &::DefWindowProcA;
        //EBhE̍쐬
        hWnd = ::CreateWindowExA(rParameter.GetExStyle(), wc.lpszClassName, NULL, rParameter.GetStyle(),
            CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, rParameter.GetParent(), rParameter.GetMenu(),
            wc.hInstance, this);
    }

    SetHandle(hWnd);  // Ô߂łB

    if (!bRegister) {
        m_fnWndProc = reinterpret_cast<WNDPROC>(WarnLess_SetWindowLongPtr(
            hWnd, GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(&CWindow::WindowProc)));
        WarnLess_SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
    }
    if (hWnd == NULL) {
        throw Win32Exception(L"EBhE̍쐬Ɏs");
    }
}

void CWindow::InnerDialogCreate(const IWindowParameter& r) {
    const CDialogParameter& rParameter = static_cast<const CDialogParameter&>(r);
    m_eMode = EModeDialog;
    //_CAO쐬[
    HWND hDlg = NativeString::IsUnicode() ?
        ::CreateDialogParamW(::GetModuleHandleW(NULL), MAKEINTRESOURCEW(rParameter.GetId()),
        rParameter.GetParent(), &CWindow::StartDialogProc, reinterpret_cast<LPARAM>(this)) :
        ::CreateDialogParamA(::GetModuleHandleA(NULL), MAKEINTRESOURCEA(rParameter.GetId()),
        rParameter.GetParent(), &CWindow::StartDialogProc, reinterpret_cast<LPARAM>(this));
    if (hDlg == NULL) {
        throw Win32Exception(L"_CAO̍쐬Ɏs");
    }
    SetHandle(hDlg); // Ô߂łB
}

void CWindow::Delete() {
    if (Exists()) {
        if (m_eMode == EModeWindow) {
            Hide();                 //Ƃ肠Bāc
            SetStyle(WS_POPUP);     //^XNo[ɃnRcȂ悤
            SetTopMost(false);      //Creatȅ(?)ɂȂƂȂ炵BB
            BOOL bRet = ::DestroyWindow(GetHandle());
            Themis_DebugAssertApi(bRet != FALSE);
            //ŃANZXo܂G[oƂƂ́A
            //bZ[W|vXbhȊODelete()ĂяoƂƁB
            //ʃXbhꍇClose()ƂɂׂB
            SetHandle(NULL);
            if (!string::IsNullOrEmpty(m_strClassName)) {
                bRet = NativeString::IsUnicode() ?
                    ::UnregisterClassW(m_strClassName.ToCharPtr(), ::GetModuleHandleW(NULL)) : 
                    ::UnregisterClassA(Encoding::GetDefault().GetBytes(m_strClassName).c_str(), ::GetModuleHandleA(NULL));
                Themis_DebugAssertApi(bRet != FALSE);
            }
        } else if (m_eMode == EModeDialog) {
            ::DestroyWindow(GetHandle());
            SetHandle(NULL);
        } else if (m_eMode == EModeDialogModal) {
            ::EndDialog(GetHandle(), result);
            SetHandle(NULL);
        } else {
            Themis_DebugTrespass();
        }
    }
}

namespace {
    set_vector<DWORD>& GetThreadIdList() {
        static set_vector<DWORD> s_oThreadIdList;
        return s_oThreadIdList;
    }
}

/// bZ[WXbh
void CWindow::OnThread(Detail::SWindowCreateParam* pParam) {
    Thread::SetName(L"WindowThread");
    Themis_DebugAssert(m_pThreadData.get() != NULL);
    Initialize(*pParam);
    // bZ[W[v
    while (m_pThreadData->threadValid && MsgPumping()) {
        {
            boost::mutex::scoped_lock lock(m_pThreadData->mutex);
            if (m_pThreadData->bWaitMsg) {
                // bZ[WIƂŃCxgN
                m_pThreadData->condition.notify_all();
                m_pThreadData->bWaitMsg = false;
            }
        }
        ::WaitMessage();
    }
    boost::mutex::scoped_lock lock(m_pThreadData->mutex);
    if (m_pThreadData->bWaitMsg) {
        m_pThreadData->condition.notify_all();
        m_pThreadData->bWaitMsg = false;
    }
    Themis_DebugAssert(!m_pThreadData->bMsgLoopValid);
    Themis_DebugAssert(!m_pThreadData->bWaitMsg);
}

namespace {
    int& GetThreadWindowCount() {
        static map_vector<DWORD, int> s_oWindowCount;
        return (*s_oWindowCount.insert(::GetCurrentThreadId(), int(0)).second).second;
    }
}

/// EBhE̍쐬Ƃ
void CWindow::Initialize(Detail::SWindowCreateParam& rParam) {
    // eXbh̃EBhEJEgP₷
    GetThreadWindowCount()++;
    try {
     // Themis_DebugAssert(!IsOwnWindowThread(::GetCurrentThreadId()));
        // RɈĎ́A̃Xbh
        // Qڈȏ̃EBhE쐬悤ƂĂƂƁB
        // WM_DESTROYŏPostQuitMessage()Ă̂ŁA
        // ꂪ悤ȏꍇ͎gȂ悤ɁB
        //
        //  Xbh̃EBhE悤ɂ̂ővɂȂ
        //
        if (m_pThreadData.get() != NULL) {
            m_pThreadData->bMsgLoopValid = true;
        }
        GetThreadIdList().insert(::GetCurrentThreadId());
        // ̍쐬
        (this->*(rParam.fnCreate))(*rParam.pParameter);
        // ANeBuɂ
        if (rParam.pParameter->IsActiveOnCreate()) {
            SetForeground();
        }
    } catch (std::exception& e) {
        GetThreadWindowCount()--;
        //쐬s
        rParam.strErrorMsg = Encoding::GetDefault().GetString(e.what());
        SetHandle(NULL);
        if (m_pThreadData.get() != NULL) {
            boost::mutex::scoped_lock lock(m_pThreadData->mutex);
            m_pThreadData->condition.notify_all();
            m_pThreadData->bMsgLoopValid = false;
        }
        GetThreadIdList().erase(::GetCurrentThreadId());
        return;
    }
    // 쐬ICxg
    if (m_pThreadData.get() != NULL) {
        boost::mutex::scoped_lock lock(m_pThreadData->mutex);
        m_pThreadData->condition.notify_all();
    }
}

LRESULT CALLBACK CWindow::StartWindowProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
// WM_NCCREATEpWindowProcB(CWindow::WindowProcł̕炷ׂ̃P`m)
//  Themis_DebugAssert(uMsg == WM_NCCREATE);    ȒOɂQ炢ł̂ۂc
  #if 0
    if (uMsg == WM_NCCREATE) {
  #else
    if (uMsg == WM_CREATE) {
  #endif
        CWindow* pWindow = static_cast<CWindow*>(reinterpret_cast<CREATESTRUCT*>(lPrm)->lpCreateParams);
        Themis_DebugAssert(pWindow != NULL);
        pWindow->SetHandle(hWnd);
        WarnLess_SetWindowLongPtr(hWnd, GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(&CWindow::WindowProc));
        WarnLess_SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(pWindow));
        return pWindow->InnerWindowProc(hWnd, uMsg, wPrm, lPrm);
    } else {
        return NativeString::IsUnicode() ?
            ::DefWindowProcW(hWnd, uMsg, wPrm, lPrm) :
            ::DefWindowProcA(hWnd, uMsg, wPrm, lPrm);
    }
}
LRESULT CALLBACK CWindow::WindowProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
    try {
        CWindow* pWindow = reinterpret_cast<CWindow*>(WarnLess_GetWindowLongPtr(hWnd, GWLP_USERDATA));
        Themis_DebugAssert(pWindow != NULL);
        if (pWindow != NULL) {
            return pWindow->InnerWindowProc(hWnd, uMsg, wPrm, lPrm);
        }
        return NativeString::IsUnicode() ?
            ::DefWindowProcW(hWnd, uMsg, wPrm, lPrm) :
            ::DefWindowProcA(hWnd, uMsg, wPrm, lPrm); 
    } catch (std::exception& e) {
        CWnd::RemoveQuitMsg();
        CWnd().MsgBox(L"G[ (1)", e);
    }
    return 0;
}

BOOL CALLBACK CWindow::StartDialogProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
// WM_INITDIALOGpWindowProcB(CWindow::DialogProcł̕炷ׂ̃P`m)
//  Themis_DebugAssert(uMsg == WM_INITDIALOG);  ȒOɉłۂH
    if (uMsg == WM_INITDIALOG) {
        CWindow* pWindow = reinterpret_cast<CWindow*>(lPrm);
        Themis_DebugAssert(pWindow != NULL);
        Themis_DebugAssert(pWindow->GetHandle() == NULL);
        pWindow->SetHandle(hWnd);
        WarnLess_SetWindowLongPtr(hWnd, DWLP_DLGPROC, reinterpret_cast<LONG_PTR>(&CWindow::DialogProc));
        WarnLess_SetWindowLongPtr(hWnd, GWLP_USERDATA, lPrm);
        return pWindow->InnerDialogProc(hWnd, uMsg, wPrm, lPrm);
    } else {
        return FALSE;
    }
}
BOOL CALLBACK CWindow::DialogProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
    try {
        CWindow* pWindow = reinterpret_cast<CWindow*>(WarnLess_GetWindowLongPtr(hWnd, GWLP_USERDATA));
        Themis_DebugAssert(pWindow != NULL);
        return (pWindow != NULL)? pWindow->InnerDialogProc(hWnd, uMsg, wPrm, lPrm): FALSE;
    } catch (std::exception& e) {
        CWnd::RemoveQuitMsg();
        CWnd().MsgBox(L"G[ (2)", e);
    }
    return FALSE;
}

LRESULT CWindow::InnerWindowProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
    Themis_DebugAssert(m_fnWndProc != NULL);
    LRESULT lTmpRet = 0;
    return InnerWindowProc(hWnd, uMsg, wPrm, lPrm, &lTmpRet) ?
        lTmpRet : m_fnWndProc(hWnd, uMsg, wPrm, lPrm);
}

BOOL CWindow::InnerDialogProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm) {
    LRESULT lTmpRet = static_cast<LRESULT>(TRUE);
    return InnerWindowProc(hWnd, uMsg, wPrm, lPrm, &lTmpRet)? static_cast<BOOL>(lTmpRet): FALSE;
}

bool CWindow::InnerWindowProc(HWND hWnd, UINT uMsg, WPARAM wPrm, LPARAM lPrm, LRESULT* pRet) {
    try {
        Themis_DebugAssert(pRet != NULL);
        CMsgHookPrms oPrms(hWnd, uMsg, wPrm, lPrm);
        oPrms.SetResult(*pRet);
        bool ret = m_pMsgHook->Dispatch(oPrms);
        *pRet = oPrms.GetResult();
        return ret;
    } catch (std::exception& e) {
        CWnd::RemoveQuitMsg();
        m_fnException(e, m_pExceptionParam);
        return false;
    }
}

bool CWindow::HookMsg(CMsgHookPrms& rPrms) {
    if (rPrms.GetMsg() == WM_NULL) {
        if (m_nCursorState != 0) {
            if (m_nCursorState == 1) {
                ShowCursor();
                m_nCursorState = 0;
            } else if (m_nCursorState == 2) {
                HideCursor();
                m_nCursorState = 0;
            }
        }
        return false;
    } else if (rPrms.GetMsg() == WM_ENDSESSION) { //&& static_cast<BOOL>(rPrms.GetWParam()) != FALSE
        Delete(); //n͑̃fXgN^ɏAR炢ĂˁcH
        return false;
    } else if (rPrms.GetMsg() == WM_DESTROY) {
        // eXbh̃EBhEJEgP炵āAȂIbZ[W
        if (m_eMode != EModeDialogModal) {
            if (--GetThreadWindowCount() <= 0) {
                PostQuitMsg();
            }
        }
        return true;
    } else if (rPrms.GetMsg() == WM_NCDESTROY) {
        WarnLess_SetWindowLongPtr(rPrms.GetWnd(), GWLP_USERDATA, NULL);
        SetHandle(NULL);
        return true;
    } else {
        return false;
    }
}

bool CWindow::IsMsgLoopValid() {
    if (::GetCurrentThreadId() == ::GetWindowThreadProcessId(GetHandle(), NULL)) {
        return MsgPumping();
    } else if (m_pThreadData.get() != NULL) {
        return m_pThreadData->bMsgLoopValid;
    } else {
        Themis_DebugWarning(L"Ȃ񂩕ςCWindow::IsMsgLoopValid()̌Ăяoł");
        return false;
    }
}

void CWindow::WaitMsg() {
    if (::GetCurrentThreadId() == ::GetWindowThreadProcessId(GetHandle(), NULL)) {
        ::WaitMessage();
    } else if (m_pThreadData.get() != NULL) {
        boost::mutex::scoped_lock lock(m_pThreadData->mutex);
        m_pThreadData->bWaitMsg = true;
     // m_pThreadData->condition.wait(lock);
        m_pThreadData->condition.timed_wait(lock, XTime::AfterSeconds(2)); //ŜߓKɖ߂鎖ɂƂ
    } else {
        Themis_DebugWarning(L"Ȃ񂩕ςCWindow::WaitMsg()̌Ăяoł");
    }
}

void CWindow::WaitForQuit() {
    Thread::SetName(L"WindowMessageLoopThread");
 // while (IsMsgLoopValid()) { WaitMsg(); }
    // ꂾƖThreadIdƂ̎擾Ɣr̂ŁACCɏĂ܂
    if (::GetCurrentThreadId() == ::GetWindowThreadProcessId(GetHandle(), NULL)) {
        while (MsgPumping()) {
            ::WaitMessage();
        }
    } else if (m_pThreadData.get() != NULL) {
        boost::xtime sleepTime = XTime::AfterSeconds(2);
        while (m_pThreadData->bMsgLoopValid) {
            boost::mutex::scoped_lock lock(m_pThreadData->mutex);
            m_pThreadData->bWaitMsg = true;
            m_pThreadData->condition.timed_wait(lock, sleepTime); //ŜߓKɖ߂鎖ɂƂ
            XTime::AddSeconds(sleepTime, 2);
        }
    } else {
        Themis_DebugWarning(L"Ȃ񂩕ςCWindow::WaitForQuit()̌Ăяoł");
    }
}

/// ĂяoꂽXbh̃L[̃bZ[WAWM_QUITtrueԂ
bool CWindow::MsgPumping() {
    MSG oMsg;
 #if 0
    while (FALSE != ::PeekMessage(&oMsg, NULL, 0, 0, PM_NOREMOVE)) {
        int ret = ::GetMessage(&oMsg, NULL, 0, 0);
        if (ret == 0 || ret == -1) {
 #else
    while (FALSE != ::PeekMessage(&oMsg, NULL, 0, 0, PM_REMOVE)) {
        if (oMsg.message == WM_QUIT) {
 #endif
            //bZ[W[vI
            GetThreadIdList().erase(::GetCurrentThreadId());
            if (m_pThreadData.get() != NULL) {
                m_pThreadData->bMsgLoopValid = false;
            }
            return false;
        }
     #if 0
        // oMsg.hwndCWindowIuWFNg擾
        CWindow* pWindow = NULL;
        if (oMsg.hwnd != NULL) {
            pWindow = reinterpret_cast<CWindow*>(WarnLess_GetWindowLongPtr(oMsg.hwnd, GWLP_USERDATA));
            if (pWindow == NULL || pWindow->GetHandle() != oMsg.hwnd) {
                Themis_DebugAssert(pWindow == NULL); // RɈEBhEIuWFNgsۂꍇ
                pWindow = this; // NULL̕낤BB
            }
        }
        if (pWindow == NULL) {
            // ܂CWindowIuWFNg蓖ĂĂȂۂꍇ
            ::TranslateMessage(&oMsg);
            ::DispatchMessage(&oMsg);
        } else {
            // ΉCWindowIuWFNgۂꍇ
            if (pWindow->GetAccel() == NULL || ::TranslateAccelerator(oMsg.hwnd, pWindow->GetAccel(), &oMsg) == 0) {
                if (pWindow->m_eMode != EModeDialog || ::IsDialogMessage(oMsg.hwnd, &oMsg) == FALSE) {
                    ::TranslateMessage(&oMsg);
                    ::DispatchMessage(&oMsg);
                }
            }
        }
     #else
        // 蔲ŁH
        if (this->GetAccel() == NULL || ::TranslateAccelerator(this->GetHandle(), this->GetAccel(), &oMsg) == 0) {
            if (this->m_eMode != EModeDialog || ::IsDialogMessage(this->GetHandle(), &oMsg) == FALSE) {
                ::TranslateMessage(&oMsg);
                ::DispatchMessage(&oMsg);
            }
        }
     #endif
    }
    return true;
}

namespace {
    static void DefaultExceptionHandler(const std::exception& e, void* pParam) {
        CWindow* pWindow = static_cast<CWindow*>(pParam);
        if (pWindow->Exists()) {
            pWindow->MsgBox(L"G[ (3)", e);
        } else {
            CWnd().MsgBox(L"G[ (3)", e);
        }
    }
}

void CWindow::SetDefaultExceptionHandler() {
    SetExceptionHandler(&DefaultExceptionHandler, this);
}

void CWindow::SetAccel(int nID) {
    HACCEL hAccel = ::LoadAccelerators(::GetModuleHandle(NULL), MAKEINTRESOURCE(nID));
    Themis_DebugAssertApi(hAccel != NULL);
    SetAccel(hAccel);
}

void CWindow::ShowCursor() {
    if (::GetCurrentThreadId() == ::GetWindowThreadProcessId(GetHandle(), NULL)) {
        CSystem::ShowCursor();
    } else {
        m_nCursorState = 1;
        PostMsg(WM_NULL, 0, 0);
    }
}

void CWindow::HideCursor() {
    if (::GetCurrentThreadId() == ::GetWindowThreadProcessId(GetHandle(), NULL)) {
        CSystem::HideCursor();
    } else {
        m_nCursorState = 2;
        PostMsg(WM_NULL, 0, 0);
    }
}

bool CWindow::IsOwnWindowThread(DWORD dwThreadId) {
    return GetThreadIdList().find(dwThreadId) != GetThreadIdList().end();
}

////////////////////////////////////////////////////////////////////////////////

bool CMenu::IsEnabled(UINT nId) {
    MENUITEMINFO mii = {sizeof(MENUITEMINFO)};
    mii.fMask = MIIM_STATE;
    ::GetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
    return (mii.fState & MFS_ENABLED) != 0;
}
void CMenu::SetEnable(UINT nId, bool bEnabled) {
    ::EnableMenuItem(GetHandle(), nId, bEnabled?MF_ENABLED:MF_GRAYED);
}
bool CMenu::ReverseEnabled(UINT nId) {
    bool bEnabled = !IsEnabled(nId);
    SetEnable(nId, bEnabled);
    return bEnabled;
}

bool CMenu::GetCheck(UINT nId) {
    MENUITEMINFO mii = {sizeof(MENUITEMINFO)};
    mii.fMask = MIIM_STATE;
    ::GetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
    return (mii.fState & MFS_CHECKED) != 0;
}
void CMenu::SetCheck(UINT nId, bool bChecked) {
    MENUITEMINFO mii = {sizeof(MENUITEMINFO)};
    mii.fMask = MIIM_STATE;
    ::GetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
    if (bChecked) {
        mii.fState |= MFS_CHECKED;
    } else {
        mii.fState &= ~MFS_CHECKED;
    }
    ::SetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
}
bool CMenu::ReverseChecked(UINT nId) {
    bool bChecked = !GetCheck(nId);
    SetCheck(nId, bChecked);
    return bChecked;
}

string CMenu::GetText(UINT nId) {
    MENUITEMINFO mii = {sizeof(MENUITEMINFO)};
    mii.fMask = MIIM_TYPE;
    ::GetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
    boost::scoped_array<char> sz(new char[mii.cch + 2]);
    mii.cch += 1;
    mii.dwTypeData = sz.get();
    ::GetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
    return Encoding::GetDefault().GetString(sz.get());
}

void CMenu::SetText(UINT nId, const string& caption) {
    std::string strCaption = Encoding::GetDefault().GetBytes(caption);
    MENUITEMINFO mii = {sizeof(MENUITEMINFO)};
    mii.fMask = MIIM_TYPE;
    mii.dwTypeData = const_cast<LPSTR>(strCaption.c_str());
    ::SetMenuItemInfo(GetHandle(), nId, FALSE, &mii);
}

void CMenu::AddItem(UINT nId, const string& strCaption) {
    ::AppendMenu(GetHandle(), MF_STRING, nId, Encoding::GetDefault().GetBytes(strCaption).c_str());
}

void CMenu::InsertItem(UINT nPos, UINT nId, const string& strCaption) {
    ::InsertMenu(GetHandle(), nPos, MF_BYPOSITION | MF_STRING, nId, Encoding::GetDefault().GetBytes(strCaption).c_str());
}

void CMenu::AddSeparator() {
    ::AppendMenu(GetHandle(), MF_SEPARATOR, 0, NULL);
}

void CMenu::InsertSeparator(UINT nPos) {
    ::InsertMenu(GetHandle(), nPos, MF_BYPOSITION | MF_SEPARATOR, 0, NULL);
}

CPopupMenu::CPopupMenu() : CMenu(::CreatePopupMenu()) {
}

CPopupMenu::~CPopupMenu() {
    ::DestroyMenu(GetHandle());
}

HMENU CPopupMenu::Release() {
    HMENU hMenu = GetHandle();
    SetHandle(NULL);
    return hMenu;
}

CResMenu::CResMenu(int nResId) : CMenu(::LoadMenu(::GetModuleHandle(NULL), MAKEINTRESOURCE(nResId))) {
}

CResMenu::~CResMenu() {
    ::DestroyMenu(GetHandle());
}

////////////////////////////////////////////////////////////////////////////////

CStatusbar::CStatusbar(HWND hWnd, CMsgHook& rHook) : m_hWindow(hWnd), m_hStatus(NULL) {
    ::InitCommonControls();
    rHook.Add(this);
    Attach();
}

CStatusbar::~CStatusbar() {
    Detach();
}

void CStatusbar::Attach() {
    Detach();

    //Xe[^Xo[쐬[
    LONG lStyle = WS_CHILD | WS_VISIBLE;    //| CCS_BOTTOM | SBARS_SIZEGRIP;
    m_hStatus = ::CreateStatusWindow(lStyle, NULL, m_hWindow,
        static_cast<UINT>(reinterpret_cast<uintptr_t>(this) ^ 0x3430053b));
        // ID͂ǂłǂH

    CRect rcSts;
    ::GetWindowRect(m_hStatus, &rcSts);
    m_nHeight = rcSts.GetHeight();

    CRect rcWnd;
    ::GetWindowRect(m_hWindow, &rcWnd);
    ::SetWindowPos(m_hWindow, NULL, 0, 0, rcWnd.GetWidth(), rcWnd.GetHeight()+m_nHeight, SWP_NOMOVE|SWP_NOOWNERZORDER);
}

void CStatusbar::Detach() {
    if (IsAttached()) {
        ::DestroyWindow(m_hStatus);
        m_hStatus = NULL;
    }
}

void CStatusbar::SetParts(const TPartsArray& r) {
    if (IsAttached()) {
        m_oParts = r;
        if (m_oParts.empty()) {
            m_oParts.push_back(make_pair(0, false));
        }
        InnerSetParts();
    }
}

void CStatusbar::SetText(int n, const string& szMsg) {
    if (IsAttached()) {
        m_oTexts[n] = Encoding::GetDefault().GetBytes(szMsg);
        PostMsg(SB_SETTEXT, (n|0), LPARAM(m_oTexts[n].c_str()));
    }
}

void CStatusbar::InnerSetParts() {
    if (IsAttached()) {
        CRect rc;
        ::GetWindowRect(m_hWindow, &rc);
        int nWidth = rc.GetWidth();

        size_t nSize = m_oParts.size();
        vector<int> anParts(nSize);
        for (size_t i=0; i < nSize; i++) {
            pair<int, bool>& r = m_oParts[i];
            anParts[i] = r.second ? (r.first) : (nWidth - r.first);
        }

        SendMsg(SB_SETPARTS, anParts.size(), LPARAM(&anParts[0]));
        m_oTexts.resize(anParts.size());
    }
}

bool CStatusbar::HookMsg(CMsgHookPrms& rPrms) {
    if (rPrms.GetMsg() == WM_SIZE) {     //eEBhẼTCYςƂɁB
        SendMsg(rPrms.GetMsg(), rPrms.GetWPrm(), rPrms.GetLPrm());
        InnerSetParts();
        return true;
    } else if (rPrms.GetMsg() == WM_DESTROY) {  //eƂ
        Detach();
        return true;
    }
    return false;
}

}
