#include "stdafx.h"
#include "Environment.h"
#include "NativeString.h"
#include "../String/Encoding.h"
#include "../Utility/Algorithm.h"
namespace Themis {

/// R}hC̎擾
string Environment::GetCommandLine() {
    return NativeString::IsUnicode() ?
        string(::GetCommandLineW()) :
        Encoding::GetDefault().GetString(::GetCommandLineA());
}

/// parseς݂̃R}hC̎擾
List<string> Environment::GetCommandLineArgs() {
    List<string> result;
    string commandLine = GetCommandLine();
    Themis_DebugOutput(L"CommandLine: " + commandLine);

    // .NETł̎dl ::CommandLineToArgvW() ɏۂB
/*
    CommandLineToArgvW has a special interpretation of backslash characters
    when they are followed by a quotation mark character ("), as follows:

        * 2n backslashes followed by a quotation mark produce n backslashes
            followed by a quotation mark.
        * (2n) + 1 backslashes followed by a quotation mark again produce n
            backslashes followed by a quotation mark.
        * n backslashes not followed by a quotation mark simply produce n
            backslashes.

    MSDNBB(L`)
*/

    // 󔒋؂TĕĂ
    for (int i = 0, n = commandLine.GetLength(); i < n; ) {
        while (i < n && Char::IsWhiteSpace(commandLine[i])) { i++; }
        if (n <= i) break;
        int first;
        if (commandLine[i] == L'"') {
            // L""ꍇ
            first = ++i;
            for (; i < n; i++) {
                if (commandLine[i] == L'"') {
                    result.push_back(commandLine.Substring(first, i - first));
                    i++;
                    goto Next;
                }
            }
        } else {
            // ̋󔒋؂ȏꍇ
            first = i;
            for (; i < n; i++) {
                if (commandLine[i] == L' ') {
                    result.push_back(commandLine.Substring(first, i - first));
                    i++;
                    goto Next;
                }
            }
        }
        Themis_DebugAssert(n <= i);
        if (first != i) {
            result.push_back(commandLine.Substring(first));
        }
        break;
    Next:;
    }
    return result;
}

string Environment::GetCurrentDirectory() {
    if (NativeString::IsUnicode()) {
        wchar_t sz[_MAX_PATH];
        DWORD dwResult = ::GetCurrentDirectoryW(_MAX_PATH, sz);
        Themis_DebugAssertApi(dwResult != 0);
        return sz;
    } else {
        char sz[_MAX_PATH];
        DWORD dwResult = ::GetCurrentDirectoryA(_MAX_PATH, sz);
        Themis_DebugAssertApi(dwResult != 0);
        return Encoding::GetDefault().GetString(sz);
    }
}

void Environment::SetCurrentDirectory(const string& path) {
    BOOL bResult = NativeString::IsUnicode() ?
        ::SetCurrentDirectoryW(path.ToCharPtr()) :
        ::SetCurrentDirectoryA(Encoding::GetDefault().GetBytes(path).c_str());
    Themis_DebugAssertApi(bResult != FALSE);
}

namespace {
    template<class T, class FN, class TS>
    static string DoExpandEnvironmentStrings(const T& str, FN* fn, TS toString) {
        typedef typename T::value_type value_type;
        //ϐ̒u(_MAX_PATHȏȃt@C͂Ƃ肠ϐ͓ĂȂƌȂ)
        DWORD dwSize = Max<DWORD>(_MAX_PATH, static_cast<DWORD>(str.length() + 1));
        boost::scoped_array<value_type> sz(new value_type[dwSize]);
        DWORD dw = (*fn)(str.c_str(), sz.get(), dwSize);
        if (dw == 0 || dwSize <= dw) { // dwSizeȂ͂OԂĂĂۂǁBB
            dwSize *= 2; // Ôߔ{ɂĂxĂ݂
            sz.reset(new value_type[dwSize]);
            dw = (*fn)(str.c_str(), sz.get(), dwSize);
            if (dw == 0 || dwSize <= dw) {
                // ܂ȂȂmՂBŒł _MAX_PATH * 2 𒴂ȂĖőɖ낤BB
                // G[ȏꍇĂǂȏꍇȂ񂶂납B
                throw Win32Exception(L"t@C '" + toString(str) + L"' ̊ϐ̒uɎs");
            }
        }
        return toString(sz.get());
    }
    struct ToStringNoConvert {
        string operator()(const string& str) const { return str; }
    };
    struct ToStringDefault {
        string operator()(const std::string& str) const { return Encoding::GetDefault().GetString(str); }
    };
}
string Environment::ExpandEnvironmentVariables(const string& name) {
    return NativeString::IsUnicode() ?
        DoExpandEnvironmentStrings(name.ToCharArray(), &::ExpandEnvironmentStringsW, ToStringNoConvert()) :
        DoExpandEnvironmentStrings(Encoding::GetDefault().GetBytes(name), &::ExpandEnvironmentStringsA, ToStringDefault());
}

}
