#include "stdafx.h"
#include "FileDialog.h"
#include "NativeString.h"
#include "Environment.h"
#include "../String/StringBuilder.h"
#include "../String/Encoding.h"
#include "../Utility/Algorithm.h"
#include "../Utility/StringUtility.h"
#include <commdlg.h>
#pragma comment(lib, "comdlg32.lib")
namespace Themis {

namespace {
    static void SetStr(LPCSTR& lp, const string& str, boost::ptr_deque<std::string>& buffer) {
        if (string::IsNullOrEmpty(str)) {
            lp = NULL;
        } else {
            buffer.push_back(new std::string(Encoding::GetDefault().GetBytes(str)));
            lp = buffer[buffer.size() - 1].c_str();
        }
    }
    static void SetStr(LPCWSTR& lp, const string& str, boost::ptr_deque<std::string>& buffer) {
        lp = string::IsNullOrEmpty(str) ? NULL : str.ToCharPtr();
    }

    /// OPENFILENAMEFileDialog̊evpeB̒lݒ
    template<class OPENFILENAME>
    static void FillOFN(FileDialog& fd, OPENFILENAME& ofn, boost::ptr_deque<std::string>& buffer) {
        Themis_DebugAssert(ofn.lStructSize == 0 && ofn.Flags == 0); // 0߂ĂƉ

      #ifdef OPENFILENAME_SIZE_VERSION_400
        ofn.lStructSize = OPENFILENAME_SIZE_VERSION_400;
      #else
        ofn.lStructSize = sizeof(OPENFILENAME);
      #endif
        ofn.Flags = OFN_ENABLESIZING | OFN_LONGNAMES;

        if (fd.IsCheckFileExists())     ofn.Flags |= OFN_FILEMUSTEXIST;
        if (fd.IsCheckPathExists())     ofn.Flags |= OFN_PATHMUSTEXIST;            
        if (!fd.IsValidateNames())      ofn.Flags |= OFN_NOVALIDATE;
        if (!fd.IsDereferenceLinks())   ofn.Flags |= OFN_NODEREFERENCELINKS;

        ofn.nFilterIndex = fd.GetFilterIndex();

        SetStr(ofn.lpstrFilter, fd.GetFilter(), buffer);
        if (fd.IsAddExtension()) { // ꂱƂł񂾂납H
            SetStr(ofn.lpstrDefExt, fd.GetDefaultExt(), buffer);
        }
     // SetStr(ofn.lpstrFile, fd.GetFileName(), buffer);
        SetStr(ofn.lpstrInitialDir, fd.GetInitialDirectory(), buffer);
        SetStr(ofn.lpstrTitle, fd.GetTitle(), buffer);

        Themis_DebugAssert(fd.IsShowHelp() == false); // 
        Themis_DebugAssert(fd.IsSupportMultiDottedExtensions() == false); // 悭񂯂ǖ

    }

    /// OPENFILENAMEOpenFileDialog̊evpeB̒lݒ
    template<class OPENFILENAME>
    static void FillOFN(OpenFileDialog& fd, OPENFILENAME& ofn, boost::ptr_deque<std::string>& buffer) {
        FillOFN(static_cast<FileDialog&>(fd), ofn, buffer);
        if (fd.IsMultiselect()) ofn.Flags |= (OFN_ALLOWMULTISELECT | OFN_EXPLORER);
        if (!fd.IsShowReadOnly()) {
            ofn.Flags |= OFN_HIDEREADONLY;
        } else {
            if (fd.IsReadOnlyChecked()) ofn.Flags |= OFN_READONLY;
        }
    }

    /// OPENFILENAMESaveFileDialog̊evpeB̒lݒ
    template<class OPENFILENAME>
    static void FillOFN(SaveFileDialog& fd, OPENFILENAME& ofn, boost::ptr_deque<std::string>& buffer) {
        FillOFN(static_cast<FileDialog&>(fd), ofn, buffer);
        if (fd.IsCreatePrompt()) ofn.Flags |= OFN_CREATEPROMPT;
        if (fd.IsOverwritePrompt()) ofn.Flags |= OFN_OVERWRITEPROMPT;
    }

    string ReplaceInvalidChar(const string& str) {
        StringBuilder builder = str;
     // builder.Replace(L'\\', L'-');
     // builder.Replace(L'/', L'-');
     // builder.Replace(L':', L'-');
        builder.Replace(L'*', L'-');
        builder.Replace(L'"', L'-');
        builder.Replace(L'?', L'-');
        builder.Replace(L'<', L'-');
        builder.Replace(L'>', L'-');
        builder.Replace(L'|', L'-');
        return builder.ToString();
    }
}

bool FileDialog::RunDialog(HWND hWndOwner) {
    string currentDirectory;
    if (IsRestoreDirectory()) {
        currentDirectory = Environment::GetCurrentDirectory();
    }

    bool result = InnerRunDialog();

    if (IsRestoreDirectory()) {
        Environment::SetCurrentDirectory(currentDirectory);
    }
    return result;
}

bool OpenFileDialog::InnerRunDialog() {
    boost::ptr_deque<std::string> buffer;
    if (NativeString::IsUnicode()) {
        OPENFILENAMEW ofn = {0};
        wchar_t buf[0x2000] = {0};
        ofn.lpstrFile = buf;
        ofn.nMaxFile = lengthof(buf);
        StrCopy(buf, lengthof(buf), ReplaceInvalidChar(GetFileName()));
        FillOFN(*this, ofn, buffer);
    } else {
        OPENFILENAMEA ofn = {0};
        char buf[0x2000] = {0};
        ofn.lpstrFile = buf;
        ofn.nMaxFile = lengthof(buf);
        StrCopy(buf, lengthof(buf), ReplaceInvalidChar(GetFileName()));
        FillOFN(*this, ofn, buffer);
    }

    Themis_DebugWarning(L" ");
    return false;
}

bool SaveFileDialog::InnerRunDialog() {

    Themis_DebugWarning(L" ");
    return false;
}

}
